#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Record module

Provides the object for a record and the string representation of the record
in a tcpdump trace file.

Record object attributes:
    Record(
        index       = Frame number
        length_inc  = Number of bytes included in trace
        length_orig = 1
        msecs       = 905320,
        seconds     = 1349747491,
        secs        = 4.165262937545776,
    )
"""
import nfstest_config as c
from baseobj import BaseObj

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"

class Record(BaseObj):
    """Record object

       Usage:
           from packet.record import Record

           x = Record()
    """
    def __str__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned.
           If set to 1 the representation of the object is condensed to display
           the frame number and the timestamp:
               '57 0.530957 '

           If set to 2 the representation of the object also includes the number
           of bytes on the wire, number of bytes captured and a little bit more
           verbose:
               'frame 57 @ 0.530957 secs, 42 bytes on wire, 42 packet bytes'
        """
        rdebug = self.debug_repr()
        if rdebug == 1:
            out = "%d %f " % (self.index, self.secs)
        elif rdebug == 2:
            out = "frame %d @ %f secs, %d bytes on wire, %d packet bytes" % (self.index, self.secs, self.length_inc, self.length_orig)
        else:
            out = BaseObj.__str__(self)
        return out




#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
IPv6 module

Decode IP version 6 layer.
"""
from ipv4 import IPv4
import nfstest_config as c
from ipv6addr import IPv6Addr
from packet.transport.tcp import TCP

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"

class IPv6(IPv4):
    """IPv6 object

       Usage:
           from packet.internet.ipv6 import IPv6

           x = IPv6(pktt, buffer)

       Object definition:

       IPv6(
           version       = int,
           traffic_class = int,
           flow_label    = int,
           total_size    = int,
           protocol      = int,
           hop_limit     = int,
           src           = IPv6Addr(),
           dst           = IPv6Addr(),
           options = string, # IPv6 options if available
           data = string,    # raw data of payload if protocol
                             # is not supported
       )
    """
    def __init__(self, pktt, data):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
           data:
               Raw packet data for this layer.
        """
        self.data = data
        ulist = self.unpack(8, 'IHBB')
        self.version       = (ulist[0] >> 28)
        self.traffic_class = (ulist[0] >> 20)&0xFF
        self.flow_label    = ulist[0]&0xFFF
        self.total_size    = ulist[1]
        self.protocol      = ulist[2]
        self.hop_limit     = ulist[3]
        self.src           = IPv6Addr(self.rawdata(16).encode('hex'))
        self.dst           = IPv6Addr(self.rawdata(16).encode('hex'))
        pktt.pkt.ip = self

        if self.protocol == 6:
            # Decode TCP
            TCP(pktt, self.data)
            del self.data
        return




#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
NFS utilities module

Provides a set of tools for testing NFS including methods for starting a packet
trace, stopping the packet trace and then open the packet trace for analysis.
It also provides a mechanism to enable NFS/RPC kernel debug and saving the
log messages for further analysis.

Furthermore, methods for finding specific NFSv4 operations within the packet
trace is also included.
"""
import os
import re
import subprocess
from host import Host
import nfstest_config as c
from packet.pktt import Pktt
from packet.nfs.nfs4_const import *

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"

class NFSUtil(Host):
    """NFSUtil object

       NFSUtil() -> New NFSUtil object

       Usage:
           from nfstest.nfs_util import NFSUtil

           # Create object for local host
           x = NFSUtil()

           # Start packet trace
           x.trace_start()

           # Stop packet trace
           x.trace_stop()

           # Open packet trace
           x.trace_open()

           # Enable NFS kernel debug
           x.nfs_debug_enable(nfsdebug='all'):

           # Stop NFS kernel debug
           x.nfs_debug_reset()
    """
    def __init__(self, **kwargs):
        """Constructor

           Initialize object's private data.

           rpcdebug:
               Set RPC kernel debug flags and save log messages [default: '']
           nfsdebug:
               Set NFS kernel debug flags and save log messages [default: '']
           dbgname:
               Base name for log messages files to create [default: 'dbgfile']
           tracename:
               Base name for trace files to create [default: 'tracefile']
           notrace:
               Debug option so a trace is not actually started [default: False]
           tcpdump:
               Tcpdump command [default: '/usr/sbin/tcpdump']
           messages:
               Location of file for system messages [default: '/var/log/messages']
           tmpdir:
               Temporary directory where trace files are created [default: '/tmp']
        """
        # Arguments
        self.rpcdebug  = kwargs.pop("rpcdebug",  '')
        self.nfsdebug  = kwargs.pop("nfsdebug",  '')
        self.dbgname   = kwargs.pop("dbgname",   'dbgfile')
        self.tracename = kwargs.pop("tracename", 'tracefile')
        self.notrace   = kwargs.pop("notrace",   False)
        self.tcpdump   = kwargs.pop("tcpdump",   c.NFSTEST_TCPDUMP)
        self.messages  = kwargs.pop("messages",  c.NFSTEST_MESSAGESLOG)
        self.tmpdir    = kwargs.pop("tmpdir",    c.NFSTEST_TMPDIR)
        self._nfsdebug = False
        Host.__init__(self)

        # Initialize object variables
        self.dbgidx = 1
        self.dbgfile = ''
        self.traceidx = 1
        self.tracefile = ''
        self.tracefiles = []
        self.clients = []
        self.traceproc = None
        self.nii_name = ''    # nii_name for the client
        self.nii_server = ''  # nii_name for the server

    def trace_start(self, tracefile=None, interface=None, capsize=None, clients=None):
        """Start trace on interface given

           tracefile:
               Name of trace file to create, default is a unique name
               created in the temporary directory using self.tracename as the
               base name.
           capsize:
               Use the -C option of tcpdump to split the trace files every
               1000000*capsize bytes. See documentation for tcpdump for more
               information
           clients:
               List of Host() objects to monitor

           Return the name of the trace file created.
        """
        self.trace_stop()
        if tracefile:
            self.tracefile = tracefile
        else:
            self.tracefile = "%s/%s_%d.cap" % (self.tmpdir, self.tracename, self.traceidx)
            self.traceidx += 1
        if not self.notrace:
            if len(self.nfsdebug) or len(self.rpcdebug):
                self.nfs_debug_enable()
            self.tracefiles.append(self.tracefile)

            if clients is None:
                clients = self.clients

            if interface is None:
                interface = self.interface

            opts = ""
            if interface is not None:
                opts += " -i %s" % interface

            if capsize:
                opts += " -C %d" % capsize

            hosts = self.ipaddr
            for cobj in clients:
                hosts += " or %s" % cobj.ipaddr

            cmd = "%s%s -s 0 -w %s host %s" % (self.tcpdump, opts, self.tracefile, hosts)
            self.run_cmd(cmd, sudo=True, dlevel='DBG2', msg="Trace start: ", wait=False)
            self.traceproc = self.process

            # Make sure tcpdump has started
            out = self.traceproc.stderr.readline()
            if re.search('not found', out):
                raise Exception(out)
        return self.tracefile

    def trace_stop(self):
        """Stop the trace started by trace_start()."""
        try:
            if self.traceproc:
                self.dprint('DBG2', "Trace stop")
                self.stop_cmd(self.traceproc)
                self.traceproc = None
            if not self.notrace and self._nfsdebug:
                self.nfs_debug_reset()
        except:
            return

    def trace_open(self, tracefile=None, **kwargs):
        """Open the trace file given or the trace file started by trace_start().

           All extra options are passed directly to the packet trace object.

           Return the packet trace object created, the packet trace object
           is also stored in the object attribute pktt.
        """
        if tracefile is None:
            tracefile = self.tracefile
        self.dprint('DBG1', "trace_open [%s]" % tracefile)
        self.pktt = Pktt(tracefile, **kwargs)
        return self.pktt

    def nfs_debug_enable(self, **kwargs):
        """Enable NFS debug messages.

           rpcdebug:
               Set RPC kernel debug flags and save log messages [default: self.rpcdebug]
           nfsdebug:
               Set NFS kernel debug flags and save log messages [default: self.nfsdebug]
           dbgfile:
               Name of log messages file to create, default is a unique name
               created in the temporary directory using self.dbgname as the
               base name.
        """
        modmsgs = {
            'nfs': kwargs.pop('nfsdebug', self.nfsdebug),
            'rpc': kwargs.pop('rpcdebug', self.rpcdebug),
        }
        dbgfile = kwargs.pop('dbgfile', None)
        if dbgfile is not None:
            self.dbgfile = dbgfile
        else:
            self.dbgfile = "%s/%s_%d.msg" % (self.tmpdir, self.dbgname, self.dbgidx)
            self.dbgidx += 1

        if modmsgs['nfs'] is None and modmsgs['rpc'] is None:
            return

        if os.path.exists(self.messages):
            fstat = os.stat(self.messages)
            self.dbgoffset = fstat.st_size
            self.dbgmode = fstat.st_mode & 0777
            for mod in modmsgs.keys():
                if len(modmsgs[mod]):
                    self._nfsdebug = True
                    cmd = "rpcdebug -v -m %s -s %s" % (mod, modmsgs[mod])
                    self.run_cmd(cmd, sudo=True, dlevel='DBG2', msg="NFS debug enable: ")

    def nfs_debug_reset(self):
        """Reset NFS debug messages."""
        for mod in ('nfs', 'rpc'):
            try:
                cmd = "rpcdebug -v -m %s -c" % mod
                self.run_cmd(cmd, sudo=True, dlevel='DBG2', msg="NFS debug reset: ")
            except:
                pass

        if self.dbgoffset != None:
            try:
                fd = None
                fdw = None
                os.system(self.sudo_cmd("chmod %o %s" % (self.dbgmode|0444, self.messages)))
                self.dprint('DBG2', "Creating log messages file [%s]" % self.dbgfile)
                fdw = open(self.dbgfile, "w")
                fd = open(self.messages, "r")
                fd.seek(self.dbgoffset)
                while True:
                    data = fd.read(self.rsize)
                    if len(data) == 0:
                        break
                    fdw.write(data)
            except Exception as e:
                raise
            finally:
                if fd:
                    fd.close()
                if fdw:
                    fdw.close()
                os.system(self.sudo_cmd("chmod %o %s" % (self.dbgmode, self.messages)))

    def find_nfs_op(self, op, ipaddr, port=None, match='', status=0, src_ipaddr=None, maxindex=None, call_only=False):
        """Find the call and its corresponding reply for the specified NFSv4
           operation going to the server specified by the ipaddr and port.
           The reply must also match the given status.

           op:
               NFS operation to find
           ipaddr:
               Destination IP address
           port:
               Destination port [default: any destination port]
           match:
               Match string to include [default: '']
           status:
               Match the status of the operation [default: 0]
           src_ipaddr:
               Source IP address [default: any IP address]
           maxindex:
               The match fails if packet index hits this limit [default: no limit]
           call_only:
               Find the call only [default: False]

           Return a tuple: (pktcall, pktreply).
        """
        mstatus = "" if status is None else "NFS.status == %d and " % status
        src = "IP.src == '%s' and " % src_ipaddr if src_ipaddr != None else ''
        dst = "IP.dst == '%s' and " % ipaddr
        if len(match):
            match += " and "
        if port != None:
            dst += "TCP.dst_port == %d and " % port
        pktcall   = None
        pktreply  = None
        while True:
            # Find request
            pktcall = self.pktt.match(src + dst + match + "NFS.argop == %d" % op, maxindex=maxindex)
            if pktcall and not call_only:
                # Find reply
                xid = pktcall.rpc.xid
                pktreply = self.pktt.match("RPC.xid == %d and %s NFS.resop == %d" % (xid, mstatus, op), maxindex=maxindex)
                if pktreply:
                    break
            else:
                break
        return (pktcall, pktreply)

    def find_open(self, filename, ipaddr, port, deleg_type=None, deleg_stateid=None, src_ipaddr=None, maxindex=None, anyclaim=False):
        """Find the call and its corresponding reply for the NFSv4 OPEN of the
           given file going to the server specified by the ipaddr and port.

           filename:
               Find open call and reply for this file
           ipaddr:
               Destination IP address
           port:
               Destination port
           deleg_type:
               Expected delegation type on reply
           deleg_stateid:
               Delegation stateid expected on call in delegate_cur_info
           src_ipaddr:
               Source IP address [default: any IP address]
           maxindex:
               The match fails if packet index hits this limit [default: no limit]
           anyclaim:
               Find open for either regular open or using delegate_cur_info

           Return a tuple: (filehandle, open_stateid, deleg_stateid).
        """
        src = "IP.src == '%s' and " % src_ipaddr if src_ipaddr is not None else ''
        dst = self.pktt.ip_tcp_dst_expr(ipaddr, port)
        if deleg_stateid is None:
            file_str = "NFS.claim.file == '%s'" % filename
        else:
            file_str = "NFS.claim.delegate_cur_info.file == '%s'" % filename
            file_str += " and NFS.claim.delegate_cur_info.delegate_stateid.other == '%s'" % self.pktt.escape(deleg_stateid)
        if anyclaim:
            file_str = "(NFS.claim.file == '%s' or NFS.claim.delegate_cur_info.file == '%s')" % (filename, filename)

        while True:
            pktcall = self.pktt.match(src + dst + " and NFS.argop == %d and %s" % (OP_OPEN, file_str), maxindex=maxindex)
            if not pktcall:
                return (None, None, None)
            xid = pktcall.rpc.xid
            open_str = "RPC.xid == %d and NFS.status == 0 and NFS.resop == %d" % (xid, OP_OPEN)
            if deleg_type is not None:
                open_str += " and NFS.delegation.delegation_type == %d" % deleg_type

            # Find OPEN reply to get filehandle of file
            pktreply = self.pktt.match(open_str, maxindex=maxindex)
            if not pktreply:
                continue
            idx = pktreply.NFSidx

            # GETFH should be the operation following the OPEN
            filehandle   = pktreply.nfs.resarray[idx + 1].object
            open_stateid = pktreply.NFSop.stateid.other
            if pktreply.NFSop.delegation.delegation_type in [OPEN_DELEGATE_READ, OPEN_DELEGATE_WRITE]:
                deleg_stateid = pktreply.NFSop.delegation.stateid.other
            else:
                deleg_stateid = None

            return (filehandle, open_stateid, deleg_stateid)

    def find_layoutget(self, filehandle):
        """Find the call and its corresponding reply for the NFSv4 LAYOUTGET
           of the given file handle going to the server specified by the
           ipaddr for self.server and port given by self.port.

           Return a tuple: (layoutget, layoutget_res, loc_body).
        """
        dst = self.pktt.ip_tcp_dst_expr(self.server_ipaddr, self.port)

        # Find LAYOUTGET request
        pkt = self.pktt.match(dst + " and NFS.argop == %d and NFS.object == '%s'" % (OP_LAYOUTGET, self.pktt.escape(filehandle)))
        if not pkt:
            return (None, None, None)
        xid = pkt.rpc.xid
        idx = pkt.NFSidx
        # The matched operation index (NFSidx) gives the PUTFH prior to the LAYOUTGET
        # since NFS.object is the last match
        layoutget = pkt.nfs.argarray[idx+1]

        # Find LAYOUTGET reply
        pkt = self.pktt.match("RPC.xid == %d and NFS.resop == %d" % (xid, OP_LAYOUTGET))
        if pkt is None:
            return (layoutget, None, None)
        layoutget_res = pkt.NFSop
        if layoutget_res.logr_status:
            return (layoutget, layoutget_res, None)
        # XXX Using first layout segment only
        layout = layoutget_res.logr_layout[0]

        # Get layout content
        loc_body = layout.lo_content.loc_body
        if layoutget.loga_layout_type == LAYOUT4_NFSV4_1_FILES:
            nfl_util = loc_body.nfl_util

            # Decode loc_body
            loc_body = {
                'type':               layoutget.loga_layout_type,
                'dense':              (nfl_util & NFL4_UFLG_DENSE > 0),
                'commit_mds':         (nfl_util & NFL4_UFLG_COMMIT_THRU_MDS > 0),
                'stripe_size':        nfl_util & NFL4_UFLG_STRIPE_UNIT_SIZE_MASK,
                'first_stripe_index': loc_body.nfl_first_stripe_index,
                'offset':             loc_body.nfl_pattern_offset,
                'filehandles':        loc_body.nfl_fh_list,
                'deviceid':           loc_body.nfl_deviceid,
                'stateid':            layoutget_res.logr_stateid.other,
                'iomode':             layout.lo_iomode,
            }
        else:
            loc_body = {
                'type':               layoutget.loga_layout_type,
                'stateid':            layoutget_res.logr_stateid.other,
                'iomode':             layout.lo_iomode,
            }

        return (layoutget, layoutget_res, loc_body)

    def find_getdeviceinfo(self, deviceid=None):
        """Find the call and its corresponding reply for the NFSv4 GETDEVICEINFO
           going to the server specified by the ipaddr for self.server and port
           given by self.port.

           deviceid:
               Look for an specific deviceid [default: any deviceid]

           Return a tuple: (pktcall, pktreply, dslist).
        """
        dslist = []
        # Find GETDEVICEINFO request and reply
        match = "NFS.gdia_device_id == '%s'" % self.pktt.escape(deviceid) if deviceid is not None else ''
        (pktcall, pktreply) = self.find_nfs_op(OP_GETDEVICEINFO, self.server_ipaddr, self.port, match=match, status=None)
        if pktreply and pktreply.nfs.status == 0:
            self.gdir_device = pktreply.NFSop.gdir_device_addr
            if self.gdir_device.da_layout_type == LAYOUT4_NFSV4_1_FILES:
                da_addr_body = self.gdir_device.da_addr_body
                self.stripe_indices = da_addr_body.nflda_stripe_indices
                multipath_ds_list = da_addr_body.nflda_multipath_ds_list

                for ds_list in multipath_ds_list:
                    for item in ds_list:
                        # Get ip address and port for DS
                        addr_list = item.na_r_addr.split('.')
                        ipaddr = '.'.join(addr_list[:4])
                        port = (int(addr_list[4])<<8) + int(addr_list[5])
                        dslist.append({'ipaddr': ipaddr, 'port': port})
            # Save device info for future reference
            self.device_info[pktcall.NFSop.gdia_device_id] = {
                'call':  pktcall,
                'reply': pktreply,
            }
        if len(dslist) > 0:
            self.dslist = dslist
        return (pktcall, pktreply, dslist)

    def find_exchange_id(self, **kwargs):
        """Find the call and its corresponding reply for the NFSv4 EXCHANGE_ID
           going to the server specified by the ipaddr and port.

           ipaddr:
               Destination IP address [default: self.server]
           port:
               Destination port [default: self.port]

           Store the callback IP/TCP expression in object attribute cb_dst

           Return a tuple: (pktcall, pktreply).
        """
        ipaddr = kwargs.pop('ipaddr', self.server_ipaddr)
        port   = kwargs.pop('port', self.port)
        # Find EXCHANGE_ID request and reply
        (pktcall, pktreply) = self.find_nfs_op(OP_EXCHANGE_ID, ipaddr, port)
        self.src_ipaddr = pktcall.ip.src
        self.src_port   = pktcall.tcp.src_port
        self.cb_dst     = self.pktt.ip_tcp_dst_expr(self.src_ipaddr, self.src_port)
        try:
            if len(pktcall.NFSop.eia_client_impl_id) > 0:
                self.nii_name = pktcall.NFSop.eia_client_impl_id[0].nii_name
        except:
            pass
        try:
            if len(pktreply.NFSop.eir_server_impl_id) > 0:
                self.nii_server = pktreply.NFSop.eir_server_impl_id[0].nii_name
        except:
            pass
        return (pktcall, pktreply)

    def find_layoutrecall(self, status=0):
        """Find NFSv4 CB_LAYOUTRECALL call and return its reply.
           The reply must also match the given status.
        """
        # Find CB_LAYOUTRECALL request
        pktcall = self.pktt.match(self.cb_dst + " and NFS.argop == %d" % OP_CB_LAYOUTRECALL)
        if pktcall:
            # Find reply
            xid = pktcall.rpc.xid
            pktreply = self.pktt.match("RPC.xid == %d and NFS.resop == %d and NFS.clorr_status == %d" % (xid, OP_CB_LAYOUTRECALL, status))
        else:
            self.test(False, "CB_LAYOUTRECALL was not found")
            return
        return pktreply

    def get_abs_offset(self, offset, ds_index=None):
        """Get real file offset given by the (read/write) offset on the given
           data server index, taking into account the type of layout
           (dense/sparse), the stripe_size, first stripe index and the number
           of filehandles. The layout information is taken from object
           attribute layout.
        """
        if ds_index is None:
            return offset
        nfhs = len(self.dslist)
        stripe_size = self.layout['stripe_size']
        first_stripe_index = self.layout['first_stripe_index']
        ds_index -= first_stripe_index;
        if ds_index < 0:
            ds_index += nfhs
        # Get real file offset given by the read/write offset to the given DS index
        if self.layout['dense']:
            # Dense layout
            n = int(offset / stripe_size)
            r = offset % stripe_size
            file_offset = (n*nfhs + ds_index)*stripe_size + r
        else:
            # Sparse layout
            file_offset = offset
        return file_offset

    def get_filehandle(self, ds_index):
        """Return filehandle from the layout list of filehandles."""
        if len(self.layout['filehandles']) > 1:
            filehandle = self.layout['filehandles'][ds_index]
        else:
            filehandle = self.layout['filehandles'][0]
        return filehandle

    def verify_stripe(self, offset, size, ds_index):
        """Verify if read/write is sent to the correct data server according
           to stripe size, first stripe index and the number of filehandles.
           The layout information is taken from object attribute layout.

           offset:
               Real file offset
           size:
               I/O size
           ds_index:
               Data server index

           Return True if stripe is correctly verified, False otherwise.
        """
        nfhs = len(self.dslist)
        if self.layout is None or ds_index is None:
            return False
        stripe_size = self.layout['stripe_size']
        first_stripe_index = self.layout['first_stripe_index']
        n = int(offset / stripe_size)
        m = int((offset + size - 1) / stripe_size)
        idx = n % nfhs
        ds_index -= first_stripe_index;
        if ds_index < 0:
            ds_index += nfhs
        return n == m and idx == ds_index

    def verify_create_session(self, ipaddr, port, ds=False, nocreate=False, ds_index=None, exchid_status=0, cs_status=0):
        """Verify initial connection to the metadata server(MDS)/data server(DS).
           Verify if EXCHANGE_ID, CREATE_SESSION, RECLAIM_COMPLETE,
           GETATTR asking for FATTR4_LEASE_TIME, and GETATTR asking for
           FATTR4_FS_LAYOUT_TYPE are all sent or not to the server.

           ipaddr:
               Destination IP address of MDS or DS
           port:
               Destination port number of MDS or DS
           ds:
               True if ipaddr/port defines a DS, otherwise MDS [default: False]
           nocreate:
               True if expecting the client NOT to send EXCHANGE_ID,
               CREATE_SESSION, and RECLAIM_COMPLETE. Otherwise, verify all
               these operations are sent by the client [default: False]
           ds_index:
               DS index used for displaying purposes only [default: None]
           exchid_status:
               Expected status for EXCHANGE_ID [default: 0]
           cs_status:
               Expected status for CREATE_SESSION [default: 0]

           Return the sessionid and it is also stored in the object
           attribute sessionid.
        """
        self.sessionid = None
        if ds:
            pnfs_use_flag = EXCHGID4_FLAG_USE_PNFS_DS
            server_type = "DS"
            if ds_index is not None:
                server_type += "(%d)" % ds_index
        else:
            pnfs_use_flag = EXCHGID4_FLAG_USE_PNFS_MDS
            server_type = "MDS"

        dsmds = ""
        if ds_index != None and ipaddr == self.server_ipaddr and port == self.port:
            # DS == MDS, client does not connect to DS, it has a connection already
            nocreate = True
            dsmds = " since DS == MDS"

        # Find EXCHANGE_ID request and reply
        (pktcall, pktreply) = self.find_nfs_op(OP_EXCHANGE_ID, ipaddr, port, status=exchid_status)
        if nocreate:
            self.test(not pktcall, "EXCHANGE_ID should not be sent to %s%s" % (server_type, dsmds))
        else:
            self.test(pktcall, "EXCHANGE_ID should be sent to %s" % server_type)
            if pktreply:
                if exchid_status:
                    self.test(pktreply.NFSop.eir_status == exchid_status, "EXCHANGE_ID reply should return %s(%d)" % (nfsstat4[exchid_status], exchid_status))
                    return
                else:
                    eir_flags = pktreply.NFSop.eir_flags
                    if len(pktreply.NFSop.eir_server_impl_id) > 0:
                        self.nii_name = pktreply.NFSop.eir_server_impl_id[0].nii_name
                    self.test(eir_flags & pnfs_use_flag != 0, "EXCHGID4_FLAG_USE_PNFS_%s should be set" % server_type, terminate=True)
                    if not ds:
                        # Check for invalid combination of eir flags
                        self.test(eir_flags & EXCHGID4_FLAG_USE_NON_PNFS == 0, "EXCHGID4_FLAG_USE_NON_PNFS should not be set")
            else:
                self.test(False, "EXCHANGE_ID reply was not found")

        # Find CREATE_SESSION request
        (pktcall, pktreply) = self.find_nfs_op(OP_CREATE_SESSION, ipaddr, port, status=cs_status)
        if nocreate:
            self.test(not pktcall, "CREATE_SESSION should not be sent to %s%s" % (server_type, dsmds))
        else:
            self.test(pktcall, "CREATE_SESSION should be sent to %s" % server_type)
            if pktreply:
                if cs_status:
                    self.test(pktreply.NFSop.csr_status == cs_status, "CREATE_SESSION reply should return %s(%d)" % (nfsstat4[cs_status], cs_status))
                    return
                else:
                    # Save the session id
                    self.sessionid = pktreply.NFSop.csr_sessionid
                    # Save the max response size
                    self.ca_maxrespsz = pktreply.NFSop.csr_fore_chan_attrs.ca_maxresponsesize

                    slotid = 0
                    test_seq = True
                    save_index = self.pktt.index
                    while True:
                        # Find first SEQUENCE requests per slot id
                        (pktcall, pktreply) = self.find_nfs_op(OP_SEQUENCE, ipaddr, port, call_only=True, match="NFS.sa_slotid == %d" % slotid)
                        if pktcall is None:
                            break
                        slotid += 1
                        if pktcall.NFSop.sa_sequenceid != 1:
                            test_seq = False
                            break
                    if slotid > 0:
                        self.test(test_seq, "SEQUENCE request should start with a sequence id of 1")
                    else:
                        self.test(False, "SEQUENCE request was not found")
                    self.pktt.rewind(save_index)
            elif pktcall:
                self.test(False, "CREATE_SESSION reply was not found")

        # Find RECLAIM_COMPLETE request
        (pktcall, pktreply) = self.find_nfs_op(OP_RECLAIM_COMPLETE, ipaddr, port, status=None)
        if nocreate:
            self.test(not pktcall, "RECLAIM_COMPLETE should not be sent to %s%s" % (server_type, dsmds))
        else:
            self.test(pktcall, "RECLAIM_COMPLETE should be sent to %s" % server_type)

        if not ds:
            # Find packet having a GETATTR asking for FATTR4_LEASE_TIME(bit 10)
            attrmatch = "NFS.attr_request & %s != 0" % hex(1 << FATTR4_LEASE_TIME)
            (pktcall, pktreply) = self.find_nfs_op(OP_GETATTR, self.server_ipaddr, self.port, match=attrmatch)
            self.test(pktcall, "GETATTR should be sent to %s asking for FATTR4_LEASE_TIME" % server_type)
            if pktreply:
                lease_time = pktreply.NFSop.obj_attributes[FATTR4_LEASE_TIME]
                self.test(lease_time > 0, "NFS server should return lease time(%d) > 0" % lease_time)
            elif pktcall:
                self.test(False, "GETATTR reply was not found")

            # Find packet having a GETATTR asking for FATTR4_SUPPORTED_ATTRS(bit 0)
            attrmatch = "NFS.attr_request & %s != 0" % hex(1 << FATTR4_SUPPORTED_ATTRS)
            (pktcall, pktreply) = self.find_nfs_op(OP_GETATTR, self.server_ipaddr, self.port, match=attrmatch)
            self.test(pktcall, "GETATTR should be sent to %s asking for FATTR4_SUPPORTED_ATTRS" % server_type)
            if pktreply:
                supported_attrs = pktreply.NFSop.obj_attributes[FATTR4_SUPPORTED_ATTRS]
                fslt_supported = supported_attrs & (1<<FATTR4_FS_LAYOUT_TYPE) != 0
                self.test(fslt_supported, "NFS server should support file type layouts (LAYOUT4_NFSV4_1_FILES)")
            elif pktcall:
                self.test(False, "GETATTR reply was not found")

            # Find packet having a GETATTR asking for FATTR4_FS_LAYOUT_TYPE(bit 62)
            attrmatch = "NFS.attr_request & %s != 0" % hex(1 << FATTR4_FS_LAYOUT_TYPE)
            (pktcall, pktreply) = self.find_nfs_op(OP_GETATTR, self.server_ipaddr, self.port, match=attrmatch)
            self.test(pktcall, "GETATTR should be sent to %s asking for FATTR4_FS_LAYOUT_TYPE" % server_type)
            if pktreply:
                # Get list of fs layout types supported by the server
                fs_layout_types = pktreply.NFSop.obj_attributes[FATTR4_FS_LAYOUT_TYPE]
                self.test(LAYOUT4_NFSV4_1_FILES in fs_layout_types, "NFS server should return LAYOUT4_NFSV4_1_FILES in fs_layout_types")
            elif pktcall:
                self.test(False, "GETATTR reply was not found")
        return self.sessionid

    def verify_layoutget(self, filehandle, iomode, riomode=None, status=0, offset=None, length=None):
        """Verify the client sends a LAYOUTGET for the given file handle.

           filehandle:
               Find LAYOUTGET for this file handle
           iomode:
               Expected I/O mode for LAYOUTGET call
           riomode:
               Expected I/O mode for LAYOUTGET reply if specified, else verify
               reply I/O mode is equal to call I/O mode if iomode == 2.
               If iomode == 1, the reply I/O mode could be equal to 1 or 2
           status:
               Expected status for LAYOUTGET reply [default: 0]
           offset:
               Expected layout range for LAYOUTGET reply [default: None]
           length:
               Expected layout range for LAYOUTGET reply [default: None]

           If both offset and length are not given, verify LAYOUTGET reply
           should be a full layout [0, NFS4_UINT64_MAX]. If only one is
           provided the following defaults are used: offset = 0,
           length = NFS4_UINT64_MAX.

           Layout information is stored in the object attribute layout.

           Return a tuple: (layoutget, layoutget_res, loc_body).
        """
        # Find LAYOUTGET for given filehandle
        layoutget, layoutget_res, loc_body = self.find_layoutget(filehandle)
        if layoutget is None:
            self.layout = None
            return (None, None, None)

        # Test layout type
        self.test(layoutget.loga_layout_type == LAYOUT4_NFSV4_1_FILES, "LAYOUTGET layout type should be LAYOUT4_NFSV4_1_FILES")

        # Test iomode
        self.test(layoutget.loga_iomode == iomode, "LAYOUTGET iomode should be %s" % self.iomode_str(iomode))

        # Test for full file layout
        self.test(layoutget.loga_offset == 0 and layoutget.loga_length == NFS4_UINT64_MAX, "LAYOUTGET should ask for full file layout")

        if layoutget_res is None:
            self.test(False, "LAYOUTGET reply should be returned")
            return (layoutget, None, None)

        if status:
            self.test(layoutget_res.logr_status == status, "LAYOUTGET reply should return error %s(%d)" % (nfsstat4[status], status))
            return (layoutget, layoutget_res, None)
        elif layoutget_res.logr_status:
            self.test(False, "LAYOUTGET reply returned %s(%d)" % (nfsstat4[layoutget_res.logr_status], layoutget_res.logr_status))
            return (layoutget, layoutget_res, None)

        # Get layout from reply
        layout = layoutget_res.logr_layout[0]

        # Test LAYOUTGET reply for correct layout type
        self.test(layout.lo_content.loc_type == LAYOUT4_NFSV4_1_FILES, "LAYOUTGET reply layout type should be LAYOUT4_NFSV4_1_FILES")

        # Test LAYOUTGET reply for correct iomode
        if riomode is not None:
            self.test(layout.lo_iomode == riomode, "LAYOUTGET reply iomode is %s when asking for a %s layout" % (self.iomode_str(riomode), self.iomode_str(iomode)))
        elif iomode == 1 and layoutget.loga_iomode in [LAYOUTIOMODE4_READ, LAYOUTIOMODE4_RW]:
            self.test(True, "LAYOUTGET reply iomode is %s when asking for a IOMODE_READ layout" % self.iomode_str(layoutget.loga_iomode))
        else:
            self.test(layout.lo_iomode == iomode, "LAYOUTGET reply iomode should be %s" % self.iomode_str(iomode))

        if offset is None and length is None:
            # Test LAYOUTGET reply for full file layout
            self.test(layout.lo_offset == 0 and layout.lo_length == NFS4_UINT64_MAX, "LAYOUTGET reply should be full file layout")
        else:
            # Test LAYOUTGET reply for correct layout range
            if offset is None:
                offset = 0
            if length is None:
                length = NFS4_UINT64_MAX
            self.test(layout.lo_offset == offset and layout.lo_length == length, "LAYOUTGET reply should be: (offset=%d, length=%d)" % (offset, length))

        # Return layout
        self.layout = loc_body
        self.layout['return_on_close'] = layoutget_res.logr_return_on_close
        return (layoutget, layoutget_res, loc_body)

    def verify_io(self, iomode, stateid, ipaddr=None, port=None, src_ipaddr=None, filehandle=None, ds_index=None, init=False, maxindex=None, pattern=None):
        """Verify I/O is sent to the server specified by the ipaddr and port.

           iomode:
               Verify reads (iomode == 1) or writes (iomode == 2)
           stateid:
               Expected stateid to use in all I/O requests
           ipaddr:
               Destination IP address of MDS or DS
               [default: do not match destination]
           port:
               Destination port number of MDS or DS
               [default: do not match destination port]
           src_ipaddr:
               Source IP address of request
               [default: do not match source]
           filehandle:
               Find I/O for this file handle. This option is used when
               verifying I/O sent to the MDS
               [default: use filehandle given by ds_index]
           ds_index:
               Data server index. This option is used when verifying I/O sent
               to the DS -- filehandle is taken from x.layout for this index
               [default: None]
           init:
               Initialized test variables [default: False]
           maxindex:
               The match fails if packet index hits this limit [default: no limit]
           pattern:
               Data pattern to compare [default: default data pattern]

           Return the number of I/O operations sent to the server.
        """
        if filehandle is None:
            filehandle = self.get_filehandle(ds_index)
        src = "IP.src == '%s' and " % src_ipaddr if src_ipaddr != None else ''
        dst = ''
        if ipaddr != None:
            dst = "IP.dst == '%s' and " % ipaddr
            if port != None:
                dst += "TCP.dst_port == %d and " % port
        fh = "NFS.object == '%s'" % self.pktt.escape(filehandle)
        save_index = self.pktt.index
        xids = []
        offsets = {}
        good_pattern = 0
        bad_pattern = 0
        self.test_offsets = []
        if init:
            self.test_seqid   = True
            self.test_stateid = True
            self.test_pattern = True
            self.test_stripe  = True
            self.test_verf    = True
            self.need_commit  = False
            self.need_lcommit = False
            self.stateid      = None
            self.max_iosize   = 0

        # Get I/O type: iomode == 1 (READ), else (WRITE)
        io_op = OP_READ if iomode == LAYOUTIOMODE4_READ else OP_WRITE

        # Find all I/O requests for MDS or current DS
        while True:
            # Find I/O request
            pkt = self.pktt.match(src + dst + fh + " and NFS.argop == %d" % io_op, maxindex=maxindex)
            if not pkt:
                break
            xids.append(pkt.rpc.xid)
            nfsop = pkt.NFSop
            self.test_offsets.append(nfsop.offset)
            if iomode == LAYOUTIOMODE4_READ:
                offsets[pkt.rpc.xid] = nfsop.offset

            if nfsop.stateid.seqid != 0:
                self.test_seqid = False
            if nfsop.stateid.other != stateid:
                self.test_stateid = False
            self.stateid = nfsop.stateid.other

            # Get real file offset
            file_offset = self.get_abs_offset(nfsop.offset, ds_index)

            if iomode == LAYOUTIOMODE4_READ:
                size = nfsop.count
            else:
                data = self.data_pattern(file_offset, len(nfsop.data), pattern=pattern)
                if data != nfsop.data:
                    bad_pattern += 1
                else:
                    good_pattern += 1
                size = len(nfsop.data)
            if self.max_iosize < size:
                self.max_iosize = size

            # Check if I/O is sent to the MDS or correct DS according to stripe size
            if ds_index is not None and not self.verify_stripe(file_offset, size, ds_index):
                self.test_stripe = False

        # Rewind trace file to saved packet index
        self.pktt.rewind(save_index)

        if iomode == LAYOUTIOMODE4_RW:
            self.dprint('DBG7', "WRITE bad/good pattern %d/%d" % (bad_pattern, good_pattern))
            if good_pattern == 0 or float(bad_pattern)/good_pattern >= 0.25:
                self.test_pattern = False
            elif bad_pattern > 0:
                self.warning("Some WRITE packets were not capture properly")

        if len(xids) == 0:
            return 0

        # Find all I/O replies for MDS or current DS
        while True:
            # Find I/O reply
            pkt = self.pktt.match("NFS.resop == %d" % io_op, maxindex=maxindex)
            if not pkt:
                break
            xid = pkt.rpc.xid
            if xid in xids:
                xids.remove(xid)
                nfsop = pkt.NFSop

                if iomode == LAYOUTIOMODE4_READ:
                    offset = offsets[xid]

                    # Get real file offset
                    file_offset = self.get_abs_offset(offset, ds_index)

                    data = self.data_pattern(file_offset, len(nfsop.data), pattern=pattern)
                    if data != nfsop.data:
                        bad_pattern += 1
                    else:
                        good_pattern += 1
                else:
                    if nfsop.committed < FILE_SYNC4:
                        # Need layout commit if reply is not FILE_SYNC4
                        self.need_lcommit = True
                    if nfsop.committed == UNSTABLE4:
                        # Need commit if reply is UNSTABLE4
                        self.need_commit = True
                    if self.writeverf is None:
                        self.writeverf = nfsop.writeverf
                    if self.writeverf != nfsop.writeverf:
                        self.test_verf = False

                if len(xids) == 0:
                    break
        nops = good_pattern + bad_pattern

        if iomode == LAYOUTIOMODE4_READ:
            self.dprint('DBG7', "READ bad/good pattern %d/%d" % (bad_pattern, good_pattern))
            if good_pattern == 0 or float(bad_pattern)/good_pattern >= 0.25:
                self.test_pattern = False
            elif bad_pattern > 0:
                self.warning("Some READ packets were not capture properly")

        if len(xids) > 0:
            self.warning("Could not find all replies to %s" % ('READ' if iomode == LAYOUTIOMODE4_READ else 'WRITE'))

        return nops

    def verify_commit(self, ipaddr, port, filehandle, init=False):
        """Verify commits are properly sent to the server specified by the
           given ipaddr and port.

           ipaddr:
               Destination IP address of MDS or DS
           port:
               Destination port number of MDS or DS
           filehandle:
               Find commits for this file handle
           init:
               Initialized test variables [default: False]

           Return the number of commits sent to the server.
        """
        dst = self.pktt.ip_tcp_dst_expr(ipaddr, port)
        fh = "NFS.object == '%s'" % self.pktt.escape(filehandle)
        save_index = self.pktt.index
        xids = []
        if init:
            self.test_commit_full = True
            self.test_no_commit   = False
            self.test_commit_verf = True

        while True:
            # Find COMMIT request for current DS
            pkt = self.pktt.match(dst + " and " + fh + " and NFS.argop == %d" % OP_COMMIT)
            if not pkt:
                break
            xids.append(pkt.rpc.xid)
            nfscommit = pkt.NFSop
            if nfscommit.offset != 0 or nfscommit.count != 0:
                self.test_commit_full = False

        ncommits = len(xids)
        if ncommits == 0:
            # No COMMIT was found
            self.test_no_commit = True
            return 0

        # Rewind trace file to saved packet index
        self.pktt.rewind(save_index)
        while True:
            # Find COMMIT reply for current DS
            pkt = self.pktt.match("NFS.resop == %d" % OP_COMMIT)
            if not pkt:
                break
            if pkt.rpc.xid in xids:
                nfscommit = pkt.NFSop
                if self.writeverf != nfscommit.writeverf:
                    self.test_commit_verf = False

        return ncommits

    def verify_layoutcommit(self, filehandle, filesize):
        """Verify layoutcommit is properly sent to the server specified by
           the ipaddr for self.server and port given by self.port.
           Verify a GETATTR asking for file size is sent within the same
           compound as the LAYOUTCOMMIT.
           Verify GETATTR returns correct size for the file.

           filehandle:
               Find layoutcommit for this file handle
           filesize:
               Expected size of file
        """
        dst = self.pktt.ip_tcp_dst_expr(self.server_ipaddr, self.port)
        fh = "NFS.object == '%s'" % self.pktt.escape(filehandle)

        # Find LAYOUTCOMMIT request
        pkt = self.pktt.match(dst + " and " + fh + " and NFS.argop == %d" % OP_LAYOUTCOMMIT)
        if self.layout['commit_mds']:
            self.test(not pkt, "LAYOUTCOMMIT should not be sent to MDS when NFL4_UFLG_COMMIT_THRU_MDS is set")
        else:
            if self.need_lcommit:
                self.test(pkt, "LAYOUTCOMMIT should be sent to MDS when NFL4_UFLG_COMMIT_THRU_MDS is not set")
            else:
                self.test(not pkt, "LAYOUTCOMMIT should not be sent to MDS (FILE_SYNC4)")
            if not pkt:
                return

            xid = pkt.rpc.xid
            layoutcommit = pkt.NFSop
            range_expr = layoutcommit.loca_offset == 0 and layoutcommit.loca_length in (filesize, NFS4_UINT64_MAX)
            self.test(range_expr, "LAYOUTCOMMIT should be sent to MDS with correct file range")
            self.test(layoutcommit.loca_stateid.other == self.layout['stateid'], "LAYOUTCOMMIT should use the layout stateid")
            self.test(layoutcommit.loca_last_write_offset.no_newoffset, "LAYOUTCOMMIT new offset should be set")
            self.test(layoutcommit.loca_last_write_offset.no_offset == (filesize - 1),
                      "LAYOUTCOMMIT last write offset (%d) should be one less than the file size (%d)" % (layoutcommit.loca_last_write_offset.no_offset, filesize))
            self.test(layoutcommit.loca_layoutupdate.lou_type == LAYOUT4_NFSV4_1_FILES, "LAYOUTCOMMIT layout type should be LAYOUT4_NFSV4_1_FILES")
            self.test(len(layoutcommit.loca_layoutupdate.lou_body) == 0, "LAYOUTCOMMIT layout update field should be empty for LAYOUT4_NFSV4_1_FILES")

            # Verify a GETATTR asking for file size is sent with LAYOUTCOMMIT
            idx = pkt.NFSidx
            getattr_arg = pkt.nfs.argarray[idx+1]
            self.test(getattr_arg.attr_request & (1 << FATTR4_SIZE), "GETATTR asking for file size is sent within LAYOUTCOMMIT compound")

            # Find LAYOUTCOMMIT reply
            pkt = self.pktt.match("RPC.xid == %d and NFS.resop == %d" % (xid, OP_LAYOUTCOMMIT))
            layoutcommit = pkt.NFSop
            if layoutcommit.locr_newsize.ns_sizechanged:
                self.test(True, "LAYOUTCOMMIT reply file size changed should be set")
                ns_size = layoutcommit.locr_newsize.ns_size
                if ns_size == filesize:
                    self.test(True, "LAYOUTCOMMIT reply file size should be correct (%d)" % ns_size)
                else:
                    self.warning("LAYOUTCOMMIT reply file size is not correct (%d)" % ns_size)
            else:
                self.test(True, "LAYOUTCOMMIT reply file size changed is not set (ERRATA)")

            # Verify GETATTR returns correct file size
            idx = pkt.NFSidx
            getattr_res = pkt.nfs.resarray[idx+1]
            self.test(getattr_res.obj_attributes[FATTR4_SIZE] == filesize, "GETATTR should return correct file size within LAYOUTCOMMIT compound")
        return

    @staticmethod
    def iomode_str(iomode):
        """Return a string representation of iomode.
           This could be run as an instance or class method.
        """
        if layoutiomode4.get(iomode):
            return layoutiomode4[iomode]
        else:
            return str(iomode)

    @staticmethod
    def bitmap_str(bitmap, count, bmap, blist):
        """Return the string representation of bitmap.

           bitmap:
               Bitmap to convert
           count:
               Number of occurrences of bitmap
           bmap:
               Dictionary mapping the bits to strings
           blist:
               List of all possible bit combinations
        """
        # Get number of instances of bitmap
        cnt = 0
        for item in blist:
            if bitmap & item == bitmap:
                cnt += 1
        plist = []
        bit = max(bmap.keys())

        # Convert bitmap to a string
        while bit > 0:
            if bitmap & bit:
                plist.append(bmap[bit])
            bit = bit >> 1
        if cnt == count:
            return " & ".join(plist)
        return




#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Pkt module

Provides the object for a packet and the string representation of the packet.
This object has an attribute for each of the layers in the packet so each layer
can be accessed directly instead of going through each layer. To access the nfs
layer object you can use 'x.nfs' instead of using 'x.ethernet.ip.tcp.rpc.nfs'
which would very cumbersome to use. Also, since NFS can be used with either
TCP or UDP it would be harder to to access the nfs object independently or
the protocol.

Packet object attributes:
    Pkt(
        record   = Record information (frame number, etc.)
        ethernet = ETHERNET II (RFC 894) object
        ip       = IPv4 object
        tcp      = TCP object
        rpc      = RPC object
        nfs      = NFS object
    )
"""
import nfstest_config as c
from baseobj import BaseObj
from packet.nfs.nfs4_type import COMPOUND4args,COMPOUND4res
from packet.nfs.nfs4_const import nfs_opnum4,nfsstat4

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"

# The order in which to display all layers in the packet
_PKT_layers = ['record', 'ethernet', 'ip', 'tcp', 'udp', 'rpc', 'nfs']
# Required layers for debug_repr(1)
_PKT_rlayers = ['record', 'ip']
# Packet layers to display as debug_repr(2) for debug_repr(1) if last layer
_PKT_mlayers = ['record', 'ethernet', 'ip']
_maxlen = len(max(_PKT_layers, key=len))

class Pkt(BaseObj):
    """Packet object

       Usage:
           from packet.pkt import Pkt

           x = Pkt()
    """
    def __str__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned.
           If set to 1 the representation of is condensed into a single line.
           It contains, the frame number, IP source and destination and/or the
           last layer:
               '1 0.386615 192.168.0.62 -> 192.168.0.17 TCP 2049 -> 708, seq: 3395733180, ack: 3294169773, ACK,SYN'
               '5 0.530957 00:0c:29:54:09:ef -> ff:ff:ff:ff:ff:ff, type: 0x806'
               '19 0.434370 192.168.0.17 -> 192.168.0.62 NFS v4 COMPOUND4 call  SEQUENCE;PUTFH;GETATTR'

           If set to 2 the representation of the object is a line for each layer:
               'Pkt(
                    RECORD:   frame 19 @ 0.434370 secs, 238 bytes on wire, 238 bytes captured
                    ETHERNET: 00:0c:29:54:09:ef -> e4:ce:8f:58:9f:f4, type: 0x800(IPv4)
                    IP:       192.168.0.17 -> 192.168.0.62, protocol: 6(TCP), len: 224
                    TCP:      src port 708 -> dst port 2049, seq: 3294170673, ack: 3395734137, len: 172, flags: ACK,PSH
                    RPC:      CALL(0), program: 100003, version: 4, procedure: 1, xid: 0x1437d3d5
                    NFS:      COMPOUND4args(tag='', minorversion=1, argarray=[nfs_argop4(argop=OP_SEQUENCE, ...), ...])
                )'
        """
        rdebug = self.debug_repr()
        if rdebug > 0:
            out = "Pkt(\n" if rdebug == 2 else ''
            klist = []
            for key in _PKT_layers:
                if hasattr(self, key):
                    klist.append(key)
            lastkey = klist[-1]
            for key in klist:
                value = getattr(self, key, None)
                if value != None and (rdebug > 1 or key == lastkey or key in _PKT_rlayers):
                    if rdebug == 1:
                        if key == 'nfs':
                            out += self._nfs_str(value)
                        else:
                            if key == lastkey and key in _PKT_mlayers:
                                self.debug_repr(2)
                            out += "%s" % str(value)
                            self.debug_repr(rdebug)
                    else:
                        sps = " " * (_maxlen - len(key))
                        out += "    %s:%s %s\n" % (key.upper(), sps, str(value))
            out += ")\n" if rdebug == 2 else ""
        else:
            out = BaseObj.__str__(self)
        return out

    def _nfs_str(self, nfs):
        """Internal method to return a condensed string representation
           of the NFS packet.
        """
        out = 'NFS v%d ' % self.rpc.version
        if isinstance(nfs, COMPOUND4args):
            out += 'COMPOUND4 call  '
            oplist = []
            for item in nfs.argarray:
                oplist.append(nfs_opnum4[item.argop][3:])
            out += ';'.join(oplist)
        elif isinstance(nfs, COMPOUND4res):
            out += 'COMPOUND4 reply '
            oplist = []
            for item in nfs.resarray:
                oplist.append(nfs_opnum4[item.resop][3:])
            out += ';'.join(oplist)
            if nfs.status != 0:
                out += ' -> ' + nfsstat4[nfs.status]
        return out




#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
IPv4 module

Decode IP version 4 layer.
"""
import struct
import nfstest_config as c
from baseobj import BaseObj
from packet.transport.tcp import TCP
from packet.unpack import Unpack

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"

# Name of different protocols
_IP_map = {1:'ICMP', 2:'IGMP', 6:'TCP', 17:'UDP' }

class TOS(BaseObj): pass
class Flags(BaseObj): pass

class IPv4(BaseObj, Unpack):
    """IPv4 object

       Usage:
           from packet.internet.ipv4 import IPv4

           x = IPv4(pktt, buffer)

       Object definition:

       IPv4(
           version         = int,
           IHL             = int,
           header_size     = int,
           total_size      = int,
           id              = int,
           fragment_offset = int,
           TTL             = int,
           protocol        = int,
           checksum        = int,
           src             = "%d.%d.%d.%d", # source IP address
           dst             = "%d.%d.%d.%d", # destination IP address
           TOS = TOS(
               precedence    = int,
               delay         = int,
               throughput    = int,
               reliability   = int,
               monetary_cost = int,
           ),
           DSCP = int,
           ECN  = int,
           flags = Flags(
               DF = int,
               MF = int,
           )
           options = string, # IP options if available
           data = string,    # Raw data of payload if protocol
                             # is not supported
       )
    """
    def __init__(self, pktt, data):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
           data:
               Raw packet data for this layer.
        """
        self.data = data
        # Get the IP version and header length
        temp = self.unpack(1, 'B')[0]
        count = 4*(temp & 0x0F)

        # Decode IP header
        ulist = self.unpack(19, 'BHHHBBH4s4s')
        self.version         = (temp >> 4)
        self.IHL             = (temp & 0x0F)
        self.header_size     = count
        self.total_size      = ulist[1]
        self.id              = ulist[2]
        self.fragment_offset = (ulist[3] & 0x1FFF)
        self.TTL             = ulist[4]
        self.protocol        = ulist[5]
        self.checksum        = ulist[6]
        self.src             = "%d.%d.%d.%d" % struct.unpack('!4B', ulist[7])
        self.dst             = "%d.%d.%d.%d" % struct.unpack('!4B', ulist[8])
        self.TOS = TOS(
            precedence    = (ulist[0] >> 5),
            delay         = ((ulist[0] >> 4) & 0x01),
            throughput    = ((ulist[0] >> 3) & 0x01),
            reliability   = ((ulist[0] >> 2) & 0x01),
            monetary_cost = ((ulist[0] >> 1) & 0x01),
        )
        self.DSCP = (ulist[0] >> 2)
        self.ECN  = (ulist[0] & 0x03)
        self.flags = Flags(
            DF = ((ulist[3] >> 14) & 0x01),
            MF = ((ulist[3] >> 13) & 0x01),
        )
        pktt.pkt.ip = self

        if count > 20:
            # Save IP options
            osize = count - 20
            self.options = self.rawdata(osize)

        if self.protocol == 6:
            # Decode TCP
            TCP(pktt, self.data)
            del self.data
        return

    def __str__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned.
           If set to 1 the representation of the object is condensed:
               '192.168.0.20 -> 192.168.0.61 '

           If set to 2 the representation of the object also includes the
           protocol and length of payload:
               '192.168.0.20 -> 192.168.0.61, protocol: 17(UDP), len: 84'
        """
        rdebug = self.debug_repr()
        if rdebug == 1:
            out = "%s -> %s " % (self.src, self.dst)
        elif rdebug == 2:
            proto = _IP_map.get(self.protocol, None)
            proto = str(self.protocol) if proto is None else "%d(%s)" % (self.protocol, proto)
            out = "%s -> %s, protocol: %s, len: %d" % (self.src, self.dst, proto, self.total_size)
        else:
            out = BaseObj.__str__(self)
        return out




# Generated by rpcgen.py from nfs4.x on Wed Jul 11 08:30:29 2012
NFSPROC4_COMPOUND = 1
CB_NULL = 0
NFSPROC4_NULL = 0
CB_COMPOUND = 1
FALSE = 0
TRUE = 1
AUTH_NONE = 0
AUTH_SYS = 1
RPCSEC_GSS = 6
NFS4_FHSIZE = 128
NFS4_VERIFIER_SIZE = 8
NFS4_OPAQUE_LIMIT = 1024
NFS4_SESSIONID_SIZE = 16
NFS4_INT64_MAX = 0x7fffffffffffffffL
NFS4_UINT64_MAX = 0xffffffffffffffffL
NFS4_INT32_MAX = 0x7fffffffL
NFS4_UINT32_MAX = 0xffffffffL
NFS4_MAXFILELEN = 0xffffffffffffffffL
NFS4_MAXFILEOFF = 0xfffffffffffffffeL
NF4REG = 1
NF4DIR = 2
NF4BLK = 3
NF4CHR = 4
NF4LNK = 5
NF4SOCK = 6
NF4FIFO = 7
NF4ATTRDIR = 8
NF4NAMEDATTR = 9
nfs_ftype4 = {
    1 : 'NF4REG',
    2 : 'NF4DIR',
    3 : 'NF4BLK',
    4 : 'NF4CHR',
    5 : 'NF4LNK',
    6 : 'NF4SOCK',
    7 : 'NF4FIFO',
    8 : 'NF4ATTRDIR',
    9 : 'NF4NAMEDATTR',
}
NFS4_OK = 0
NFS4ERR_PERM = 1
NFS4ERR_NOENT = 2
NFS4ERR_IO = 5
NFS4ERR_NXIO = 6
NFS4ERR_ACCESS = 13
NFS4ERR_EXIST = 17
NFS4ERR_XDEV = 18
NFS4ERR_NOTDIR = 20
NFS4ERR_ISDIR = 21
NFS4ERR_INVAL = 22
NFS4ERR_FBIG = 27
NFS4ERR_NOSPC = 28
NFS4ERR_ROFS = 30
NFS4ERR_MLINK = 31
NFS4ERR_NAMETOOLONG = 63
NFS4ERR_NOTEMPTY = 66
NFS4ERR_DQUOT = 69
NFS4ERR_STALE = 70
NFS4ERR_BADHANDLE = 10001
NFS4ERR_BAD_COOKIE = 10003
NFS4ERR_NOTSUPP = 10004
NFS4ERR_TOOSMALL = 10005
NFS4ERR_SERVERFAULT = 10006
NFS4ERR_BADTYPE = 10007
NFS4ERR_DELAY = 10008
NFS4ERR_SAME = 10009
NFS4ERR_DENIED = 10010
NFS4ERR_EXPIRED = 10011
NFS4ERR_LOCKED = 10012
NFS4ERR_GRACE = 10013
NFS4ERR_FHEXPIRED = 10014
NFS4ERR_SHARE_DENIED = 10015
NFS4ERR_WRONGSEC = 10016
NFS4ERR_CLID_INUSE = 10017
NFS4ERR_RESOURCE = 10018
NFS4ERR_MOVED = 10019
NFS4ERR_NOFILEHANDLE = 10020
NFS4ERR_MINOR_VERS_MISMATCH = 10021
NFS4ERR_STALE_CLIENTID = 10022
NFS4ERR_STALE_STATEID = 10023
NFS4ERR_OLD_STATEID = 10024
NFS4ERR_BAD_STATEID = 10025
NFS4ERR_BAD_SEQID = 10026
NFS4ERR_NOT_SAME = 10027
NFS4ERR_LOCK_RANGE = 10028
NFS4ERR_SYMLINK = 10029
NFS4ERR_RESTOREFH = 10030
NFS4ERR_LEASE_MOVED = 10031
NFS4ERR_ATTRNOTSUPP = 10032
NFS4ERR_NO_GRACE = 10033
NFS4ERR_RECLAIM_BAD = 10034
NFS4ERR_RECLAIM_CONFLICT = 10035
NFS4ERR_BADXDR = 10036
NFS4ERR_LOCKS_HELD = 10037
NFS4ERR_OPENMODE = 10038
NFS4ERR_BADOWNER = 10039
NFS4ERR_BADCHAR = 10040
NFS4ERR_BADNAME = 10041
NFS4ERR_BAD_RANGE = 10042
NFS4ERR_LOCK_NOTSUPP = 10043
NFS4ERR_OP_ILLEGAL = 10044
NFS4ERR_DEADLOCK = 10045
NFS4ERR_FILE_OPEN = 10046
NFS4ERR_ADMIN_REVOKED = 10047
NFS4ERR_CB_PATH_DOWN = 10048
NFS4ERR_BADIOMODE = 10049
NFS4ERR_BADLAYOUT = 10050
NFS4ERR_BAD_SESSION_DIGEST = 10051
NFS4ERR_BADSESSION = 10052
NFS4ERR_BADSLOT = 10053
NFS4ERR_COMPLETE_ALREADY = 10054
NFS4ERR_CONN_NOT_BOUND_TO_SESSION = 10055
NFS4ERR_DELEG_ALREADY_WANTED = 10056
NFS4ERR_BACK_CHAN_BUSY = 10057
NFS4ERR_LAYOUTTRYLATER = 10058
NFS4ERR_LAYOUTUNAVAILABLE = 10059
NFS4ERR_NOMATCHING_LAYOUT = 10060
NFS4ERR_RECALLCONFLICT = 10061
NFS4ERR_UNKNOWN_LAYOUTTYPE = 10062
NFS4ERR_SEQ_MISORDERED = 10063
NFS4ERR_SEQUENCE_POS = 10064
NFS4ERR_REQ_TOO_BIG = 10065
NFS4ERR_REP_TOO_BIG = 10066
NFS4ERR_REP_TOO_BIG_TO_CACHE = 10067
NFS4ERR_RETRY_UNCACHED_REP = 10068
NFS4ERR_UNSAFE_COMPOUND = 10069
NFS4ERR_TOO_MANY_OPS = 10070
NFS4ERR_OP_NOT_IN_SESSION = 10071
NFS4ERR_HASH_ALG_UNSUPP = 10072
NFS4ERR_CONN_BINDING_NOT_ENFORCED = 10073
NFS4ERR_CLIENTID_BUSY = 10074
NFS4ERR_PNFS_IO_HOLE = 10075
NFS4ERR_SEQ_FALSE_RETRY = 10076
NFS4ERR_BAD_HIGH_SLOT = 10077
NFS4ERR_DEADSESSION = 10078
NFS4ERR_ENCR_ALG_UNSUPP = 10079
NFS4ERR_PNFS_NO_LAYOUT = 10080
NFS4ERR_NOT_ONLY_OP = 10081
NFS4ERR_WRONG_CRED = 10082
NFS4ERR_WRONG_TYPE = 10083
NFS4ERR_DIRDELEG_UNAVAIL = 10084
NFS4ERR_REJECT_DELEG = 10085
NFS4ERR_RETURNCONFLICT = 10086
NFS4ERR_DELEG_REVOKED = 10087
nfsstat4 = {
    0 : 'NFS4_OK',
    1 : 'NFS4ERR_PERM',
    2 : 'NFS4ERR_NOENT',
    5 : 'NFS4ERR_IO',
    6 : 'NFS4ERR_NXIO',
    13 : 'NFS4ERR_ACCESS',
    17 : 'NFS4ERR_EXIST',
    18 : 'NFS4ERR_XDEV',
    20 : 'NFS4ERR_NOTDIR',
    21 : 'NFS4ERR_ISDIR',
    22 : 'NFS4ERR_INVAL',
    27 : 'NFS4ERR_FBIG',
    28 : 'NFS4ERR_NOSPC',
    30 : 'NFS4ERR_ROFS',
    31 : 'NFS4ERR_MLINK',
    63 : 'NFS4ERR_NAMETOOLONG',
    66 : 'NFS4ERR_NOTEMPTY',
    69 : 'NFS4ERR_DQUOT',
    70 : 'NFS4ERR_STALE',
    10001 : 'NFS4ERR_BADHANDLE',
    10003 : 'NFS4ERR_BAD_COOKIE',
    10004 : 'NFS4ERR_NOTSUPP',
    10005 : 'NFS4ERR_TOOSMALL',
    10006 : 'NFS4ERR_SERVERFAULT',
    10007 : 'NFS4ERR_BADTYPE',
    10008 : 'NFS4ERR_DELAY',
    10009 : 'NFS4ERR_SAME',
    10010 : 'NFS4ERR_DENIED',
    10011 : 'NFS4ERR_EXPIRED',
    10012 : 'NFS4ERR_LOCKED',
    10013 : 'NFS4ERR_GRACE',
    10014 : 'NFS4ERR_FHEXPIRED',
    10015 : 'NFS4ERR_SHARE_DENIED',
    10016 : 'NFS4ERR_WRONGSEC',
    10017 : 'NFS4ERR_CLID_INUSE',
    10018 : 'NFS4ERR_RESOURCE',
    10019 : 'NFS4ERR_MOVED',
    10020 : 'NFS4ERR_NOFILEHANDLE',
    10021 : 'NFS4ERR_MINOR_VERS_MISMATCH',
    10022 : 'NFS4ERR_STALE_CLIENTID',
    10023 : 'NFS4ERR_STALE_STATEID',
    10024 : 'NFS4ERR_OLD_STATEID',
    10025 : 'NFS4ERR_BAD_STATEID',
    10026 : 'NFS4ERR_BAD_SEQID',
    10027 : 'NFS4ERR_NOT_SAME',
    10028 : 'NFS4ERR_LOCK_RANGE',
    10029 : 'NFS4ERR_SYMLINK',
    10030 : 'NFS4ERR_RESTOREFH',
    10031 : 'NFS4ERR_LEASE_MOVED',
    10032 : 'NFS4ERR_ATTRNOTSUPP',
    10033 : 'NFS4ERR_NO_GRACE',
    10034 : 'NFS4ERR_RECLAIM_BAD',
    10035 : 'NFS4ERR_RECLAIM_CONFLICT',
    10036 : 'NFS4ERR_BADXDR',
    10037 : 'NFS4ERR_LOCKS_HELD',
    10038 : 'NFS4ERR_OPENMODE',
    10039 : 'NFS4ERR_BADOWNER',
    10040 : 'NFS4ERR_BADCHAR',
    10041 : 'NFS4ERR_BADNAME',
    10042 : 'NFS4ERR_BAD_RANGE',
    10043 : 'NFS4ERR_LOCK_NOTSUPP',
    10044 : 'NFS4ERR_OP_ILLEGAL',
    10045 : 'NFS4ERR_DEADLOCK',
    10046 : 'NFS4ERR_FILE_OPEN',
    10047 : 'NFS4ERR_ADMIN_REVOKED',
    10048 : 'NFS4ERR_CB_PATH_DOWN',
    10049 : 'NFS4ERR_BADIOMODE',
    10050 : 'NFS4ERR_BADLAYOUT',
    10051 : 'NFS4ERR_BAD_SESSION_DIGEST',
    10052 : 'NFS4ERR_BADSESSION',
    10053 : 'NFS4ERR_BADSLOT',
    10054 : 'NFS4ERR_COMPLETE_ALREADY',
    10055 : 'NFS4ERR_CONN_NOT_BOUND_TO_SESSION',
    10056 : 'NFS4ERR_DELEG_ALREADY_WANTED',
    10057 : 'NFS4ERR_BACK_CHAN_BUSY',
    10058 : 'NFS4ERR_LAYOUTTRYLATER',
    10059 : 'NFS4ERR_LAYOUTUNAVAILABLE',
    10060 : 'NFS4ERR_NOMATCHING_LAYOUT',
    10061 : 'NFS4ERR_RECALLCONFLICT',
    10062 : 'NFS4ERR_UNKNOWN_LAYOUTTYPE',
    10063 : 'NFS4ERR_SEQ_MISORDERED',
    10064 : 'NFS4ERR_SEQUENCE_POS',
    10065 : 'NFS4ERR_REQ_TOO_BIG',
    10066 : 'NFS4ERR_REP_TOO_BIG',
    10067 : 'NFS4ERR_REP_TOO_BIG_TO_CACHE',
    10068 : 'NFS4ERR_RETRY_UNCACHED_REP',
    10069 : 'NFS4ERR_UNSAFE_COMPOUND',
    10070 : 'NFS4ERR_TOO_MANY_OPS',
    10071 : 'NFS4ERR_OP_NOT_IN_SESSION',
    10072 : 'NFS4ERR_HASH_ALG_UNSUPP',
    10073 : 'NFS4ERR_CONN_BINDING_NOT_ENFORCED',
    10074 : 'NFS4ERR_CLIENTID_BUSY',
    10075 : 'NFS4ERR_PNFS_IO_HOLE',
    10076 : 'NFS4ERR_SEQ_FALSE_RETRY',
    10077 : 'NFS4ERR_BAD_HIGH_SLOT',
    10078 : 'NFS4ERR_DEADSESSION',
    10079 : 'NFS4ERR_ENCR_ALG_UNSUPP',
    10080 : 'NFS4ERR_PNFS_NO_LAYOUT',
    10081 : 'NFS4ERR_NOT_ONLY_OP',
    10082 : 'NFS4ERR_WRONG_CRED',
    10083 : 'NFS4ERR_WRONG_TYPE',
    10084 : 'NFS4ERR_DIRDELEG_UNAVAIL',
    10085 : 'NFS4ERR_REJECT_DELEG',
    10086 : 'NFS4ERR_RETURNCONFLICT',
    10087 : 'NFS4ERR_DELEG_REVOKED',
}
SET_TO_SERVER_TIME4 = 0
SET_TO_CLIENT_TIME4 = 1
time_how4 = {
    0 : 'SET_TO_SERVER_TIME4',
    1 : 'SET_TO_CLIENT_TIME4',
}
ACL4_SUPPORT_ALLOW_ACL = 0x00000001L
ACL4_SUPPORT_DENY_ACL = 0x00000002L
ACL4_SUPPORT_AUDIT_ACL = 0x00000004L
ACL4_SUPPORT_ALARM_ACL = 0x00000008L
ACE4_ACCESS_ALLOWED_ACE_TYPE = 0x00000000L
ACE4_ACCESS_DENIED_ACE_TYPE = 0x00000001L
ACE4_SYSTEM_AUDIT_ACE_TYPE = 0x00000002L
ACE4_SYSTEM_ALARM_ACE_TYPE = 0x00000003L
ACE4_FILE_INHERIT_ACE = 0x00000001L
ACE4_DIRECTORY_INHERIT_ACE = 0x00000002L
ACE4_NO_PROPAGATE_INHERIT_ACE = 0x00000004L
ACE4_INHERIT_ONLY_ACE = 0x00000008L
ACE4_SUCCESSFUL_ACCESS_ACE_FLAG = 0x00000010L
ACE4_FAILED_ACCESS_ACE_FLAG = 0x00000020L
ACE4_IDENTIFIER_GROUP = 0x00000040L
ACE4_INHERITED_ACE = 0x00000080L
ACE4_READ_DATA = 0x00000001L
ACE4_LIST_DIRECTORY = 0x00000001L
ACE4_WRITE_DATA = 0x00000002L
ACE4_ADD_FILE = 0x00000002L
ACE4_APPEND_DATA = 0x00000004L
ACE4_ADD_SUBDIRECTORY = 0x00000004L
ACE4_READ_NAMED_ATTRS = 0x00000008L
ACE4_WRITE_NAMED_ATTRS = 0x00000010L
ACE4_EXECUTE = 0x00000020L
ACE4_DELETE_CHILD = 0x00000040L
ACE4_READ_ATTRIBUTES = 0x00000080L
ACE4_WRITE_ATTRIBUTES = 0x00000100L
ACE4_WRITE_RETENTION = 0x00000200L
ACE4_WRITE_RETENTION_HOLD = 0x00000400L
ACE4_DELETE = 0x00010000L
ACE4_READ_ACL = 0x00020000L
ACE4_WRITE_ACL = 0x00040000L
ACE4_WRITE_OWNER = 0x00080000L
ACE4_SYNCHRONIZE = 0x00100000L
ACE4_GENERIC_READ = 0x00120081L
ACE4_GENERIC_WRITE = 0x00160106L
ACE4_GENERIC_EXECUTE = 0x001200A0L
ACL4_AUTO_INHERIT = 0x00000001L
ACL4_PROTECTED = 0x00000002L
ACL4_DEFAULTED = 0x00000004L
MODE4_SUID = 0x800
MODE4_SGID = 0x400
MODE4_SVTX = 0x200
MODE4_RUSR = 0x100
MODE4_WUSR = 0x080
MODE4_XUSR = 0x040
MODE4_RGRP = 0x020
MODE4_WGRP = 0x010
MODE4_XGRP = 0x008
MODE4_ROTH = 0x004
MODE4_WOTH = 0x002
MODE4_XOTH = 0x001
FH4_PERSISTENT = 0x00000000L
FH4_NOEXPIRE_WITH_OPEN = 0x00000001L
FH4_VOLATILE_ANY = 0x00000002L
FH4_VOL_MIGRATION = 0x00000004L
FH4_VOL_RENAME = 0x00000008L
LAYOUT4_NFSV4_1_FILES = 0x1
LAYOUT4_OSD2_OBJECTS = 0x2
LAYOUT4_BLOCK_VOLUME = 0x3
layouttype4 = {
    0x1 : 'LAYOUT4_NFSV4_1_FILES',
    0x2 : 'LAYOUT4_OSD2_OBJECTS',
    0x3 : 'LAYOUT4_BLOCK_VOLUME',
}
LAYOUTIOMODE4_READ = 1
LAYOUTIOMODE4_RW = 2
LAYOUTIOMODE4_ANY = 3
layoutiomode4 = {
    1 : 'LAYOUTIOMODE4_READ',
    2 : 'LAYOUTIOMODE4_RW',
    3 : 'LAYOUTIOMODE4_ANY',
}
NFS4_DEVICEID4_SIZE = 16
LAYOUT4_RET_REC_FILE = 1
LAYOUT4_RET_REC_FSID = 2
LAYOUT4_RET_REC_ALL = 3
LAYOUTRETURN4_FILE = LAYOUT4_RET_REC_FILE
LAYOUTRETURN4_FSID = LAYOUT4_RET_REC_FSID
LAYOUTRETURN4_ALL = LAYOUT4_RET_REC_ALL
layoutreturn_type4 = {
    LAYOUT4_RET_REC_FILE : 'LAYOUTRETURN4_FILE',
    LAYOUT4_RET_REC_FSID : 'LAYOUTRETURN4_FSID',
    LAYOUT4_RET_REC_ALL : 'LAYOUTRETURN4_ALL',
}
STATUS4_FIXED = 1
STATUS4_UPDATED = 2
STATUS4_VERSIONED = 3
STATUS4_WRITABLE = 4
STATUS4_REFERRAL = 5
fs4_status_type = {
    1 : 'STATUS4_FIXED',
    2 : 'STATUS4_UPDATED',
    3 : 'STATUS4_VERSIONED',
    4 : 'STATUS4_WRITABLE',
    5 : 'STATUS4_REFERRAL',
}
TH4_READ_SIZE = 0
TH4_WRITE_SIZE = 1
TH4_READ_IOSIZE = 2
TH4_WRITE_IOSIZE = 3
RET4_DURATION_INFINITE = 0xffffffffffffffffL
FSCHARSET_CAP4_CONTAINS_NON_UTF8 = 0x1
FSCHARSET_CAP4_ALLOWS_ONLY_UTF8 = 0x2
FATTR4_SUPPORTED_ATTRS = 0
FATTR4_TYPE = 1
FATTR4_FH_EXPIRE_TYPE = 2
FATTR4_CHANGE = 3
FATTR4_SIZE = 4
FATTR4_LINK_SUPPORT = 5
FATTR4_SYMLINK_SUPPORT = 6
FATTR4_NAMED_ATTR = 7
FATTR4_FSID = 8
FATTR4_UNIQUE_HANDLES = 9
FATTR4_LEASE_TIME = 10
FATTR4_RDATTR_ERROR = 11
FATTR4_FILEHANDLE = 19
FATTR4_SUPPATTR_EXCLCREAT = 75
FATTR4_ACL = 12
FATTR4_ACLSUPPORT = 13
FATTR4_ARCHIVE = 14
FATTR4_CANSETTIME = 15
FATTR4_CASE_INSENSITIVE = 16
FATTR4_CASE_PRESERVING = 17
FATTR4_CHOWN_RESTRICTED = 18
FATTR4_FILEID = 20
FATTR4_FILES_AVAIL = 21
FATTR4_FILES_FREE = 22
FATTR4_FILES_TOTAL = 23
FATTR4_FS_LOCATIONS = 24
FATTR4_HIDDEN = 25
FATTR4_HOMOGENEOUS = 26
FATTR4_MAXFILESIZE = 27
FATTR4_MAXLINK = 28
FATTR4_MAXNAME = 29
FATTR4_MAXREAD = 30
FATTR4_MAXWRITE = 31
FATTR4_MIMETYPE = 32
FATTR4_MODE = 33
FATTR4_NO_TRUNC = 34
FATTR4_NUMLINKS = 35
FATTR4_OWNER = 36
FATTR4_OWNER_GROUP = 37
FATTR4_QUOTA_AVAIL_HARD = 38
FATTR4_QUOTA_AVAIL_SOFT = 39
FATTR4_QUOTA_USED = 40
FATTR4_RAWDEV = 41
FATTR4_SPACE_AVAIL = 42
FATTR4_SPACE_FREE = 43
FATTR4_SPACE_TOTAL = 44
FATTR4_SPACE_USED = 45
FATTR4_SYSTEM = 46
FATTR4_TIME_ACCESS = 47
FATTR4_TIME_ACCESS_SET = 48
FATTR4_TIME_BACKUP = 49
FATTR4_TIME_CREATE = 50
FATTR4_TIME_DELTA = 51
FATTR4_TIME_METADATA = 52
FATTR4_TIME_MODIFY = 53
FATTR4_TIME_MODIFY_SET = 54
FATTR4_MOUNTED_ON_FILEID = 55
FATTR4_DIR_NOTIF_DELAY = 56
FATTR4_DIRENT_NOTIF_DELAY = 57
FATTR4_DACL = 58
FATTR4_SACL = 59
FATTR4_CHANGE_POLICY = 60
FATTR4_FS_STATUS = 61
FATTR4_FS_LAYOUT_TYPE = 62
FATTR4_LAYOUT_HINT = 63
FATTR4_LAYOUT_TYPE = 64
FATTR4_LAYOUT_BLKSIZE = 65
FATTR4_LAYOUT_ALIGNMENT = 66
FATTR4_FS_LOCATIONS_INFO = 67
FATTR4_MDSTHRESHOLD = 68
FATTR4_RETENTION_GET = 69
FATTR4_RETENTION_SET = 70
FATTR4_RETENTEVT_GET = 71
FATTR4_RETENTEVT_SET = 72
FATTR4_RETENTION_HOLD = 73
FATTR4_MODE_SET_MASKED = 74
FATTR4_FS_CHARSET_CAP = 76
READ_LT = 1
WRITE_LT = 2
READW_LT = 3
WRITEW_LT = 4
nfs_lock_type4 = {
    1 : 'READ_LT',
    2 : 'WRITE_LT',
    3 : 'READW_LT',
    4 : 'WRITEW_LT',
}
SSV4_SUBKEY_MIC_I2T = 1
SSV4_SUBKEY_MIC_T2I = 2
SSV4_SUBKEY_SEAL_I2T = 3
SSV4_SUBKEY_SEAL_T2I = 4
ssv_subkey4 = {
    1 : 'SSV4_SUBKEY_MIC_I2T',
    2 : 'SSV4_SUBKEY_MIC_T2I',
    3 : 'SSV4_SUBKEY_SEAL_I2T',
    4 : 'SSV4_SUBKEY_SEAL_T2I',
}
FSLI4BX_GFLAGS = 0
FSLI4BX_TFLAGS = 1
FSLI4BX_CLSIMUL = 2
FSLI4BX_CLHANDLE = 3
FSLI4BX_CLFILEID = 4
FSLI4BX_CLWRITEVER = 5
FSLI4BX_CLCHANGE = 6
FSLI4BX_CLREADDIR = 7
FSLI4BX_READRANK = 8
FSLI4BX_WRITERANK = 9
FSLI4BX_READORDER = 10
FSLI4BX_WRITEORDER = 11
FSLI4GF_WRITABLE = 0x01
FSLI4GF_CUR_REQ = 0x02
FSLI4GF_ABSENT = 0x04
FSLI4GF_GOING = 0x08
FSLI4GF_SPLIT = 0x10
FSLI4TF_RDMA = 0x01
FSLI4IF_VAR_SUB = 0x00000001L
NFL4_UFLG_MASK = 0x0000003FL
NFL4_UFLG_DENSE = 0x00000001L
NFL4_UFLG_COMMIT_THRU_MDS = 0x00000002L
NFL4_UFLG_STRIPE_UNIT_SIZE_MASK = 0xFFFFFFC0L
NFLH4_CARE_DENSE = NFL4_UFLG_DENSE
NFLH4_CARE_COMMIT_THRU_MDS = NFL4_UFLG_COMMIT_THRU_MDS
NFLH4_CARE_STRIPE_UNIT_SIZE = 0x00000040L
NFLH4_CARE_STRIPE_COUNT = 0x00000080L
filelayout_hint_care4 = {
    NFL4_UFLG_DENSE : 'NFLH4_CARE_DENSE',
    NFL4_UFLG_COMMIT_THRU_MDS : 'NFLH4_CARE_COMMIT_THRU_MDS',
    0x00000040L : 'NFLH4_CARE_STRIPE_UNIT_SIZE',
    0x00000080L : 'NFLH4_CARE_STRIPE_COUNT',
}
ACCESS4_READ = 0x00000001L
ACCESS4_LOOKUP = 0x00000002L
ACCESS4_MODIFY = 0x00000004L
ACCESS4_EXTEND = 0x00000008L
ACCESS4_DELETE = 0x00000010L
ACCESS4_EXECUTE = 0x00000020L
UNCHECKED4 = 0
GUARDED4 = 1
EXCLUSIVE4 = 2
EXCLUSIVE4_1 = 3
createmode4 = {
    0 : 'UNCHECKED4',
    1 : 'GUARDED4',
    2 : 'EXCLUSIVE4',
    3 : 'EXCLUSIVE4_1',
}
OPEN4_NOCREATE = 0
OPEN4_CREATE = 1
opentype4 = {
    0 : 'OPEN4_NOCREATE',
    1 : 'OPEN4_CREATE',
}
NFS_LIMIT_SIZE = 1
NFS_LIMIT_BLOCKS = 2
limit_by4 = {
    1 : 'NFS_LIMIT_SIZE',
    2 : 'NFS_LIMIT_BLOCKS',
}
OPEN4_SHARE_ACCESS_READ = 0x00000001L
OPEN4_SHARE_ACCESS_WRITE = 0x00000002L
OPEN4_SHARE_ACCESS_BOTH = 0x00000003L
OPEN4_SHARE_DENY_NONE = 0x00000000L
OPEN4_SHARE_DENY_READ = 0x00000001L
OPEN4_SHARE_DENY_WRITE = 0x00000002L
OPEN4_SHARE_DENY_BOTH = 0x00000003L
OPEN4_SHARE_ACCESS_WANT_DELEG_MASK = 0xFF00
OPEN4_SHARE_ACCESS_WANT_NO_PREFERENCE = 0x0000
OPEN4_SHARE_ACCESS_WANT_READ_DELEG = 0x0100
OPEN4_SHARE_ACCESS_WANT_WRITE_DELEG = 0x0200
OPEN4_SHARE_ACCESS_WANT_ANY_DELEG = 0x0300
OPEN4_SHARE_ACCESS_WANT_NO_DELEG = 0x0400
OPEN4_SHARE_ACCESS_WANT_CANCEL = 0x0500
OPEN4_SHARE_ACCESS_WANT_SIGNAL_DELEG_WHEN_RESRC_AVAIL = 0x10000
OPEN4_SHARE_ACCESS_WANT_PUSH_DELEG_WHEN_UNCONTENDED = 0x20000
OPEN_DELEGATE_NONE = 0
OPEN_DELEGATE_READ = 1
OPEN_DELEGATE_WRITE = 2
OPEN_DELEGATE_NONE_EXT = 3
open_delegation_type4 = {
    0 : 'OPEN_DELEGATE_NONE',
    1 : 'OPEN_DELEGATE_READ',
    2 : 'OPEN_DELEGATE_WRITE',
    3 : 'OPEN_DELEGATE_NONE_EXT',
}
CLAIM_NULL = 0
CLAIM_PREVIOUS = 1
CLAIM_DELEGATE_CUR = 2
CLAIM_DELEGATE_PREV = 3
CLAIM_FH = 4
CLAIM_DELEG_CUR_FH = 5
CLAIM_DELEG_PREV_FH = 6
open_claim_type4 = {
    0 : 'CLAIM_NULL',
    1 : 'CLAIM_PREVIOUS',
    2 : 'CLAIM_DELEGATE_CUR',
    3 : 'CLAIM_DELEGATE_PREV',
    4 : 'CLAIM_FH',
    5 : 'CLAIM_DELEG_CUR_FH',
    6 : 'CLAIM_DELEG_PREV_FH',
}
WND4_NOT_WANTED = 0
WND4_CONTENTION = 1
WND4_RESOURCE = 2
WND4_NOT_SUPP_FTYPE = 3
WND4_WRITE_DELEG_NOT_SUPP_FTYPE = 4
WND4_NOT_SUPP_UPGRADE = 5
WND4_NOT_SUPP_DOWNGRADE = 6
WND4_CANCELED = 7
WND4_IS_DIR = 8
why_no_delegation4 = {
    0 : 'WND4_NOT_WANTED',
    1 : 'WND4_CONTENTION',
    2 : 'WND4_RESOURCE',
    3 : 'WND4_NOT_SUPP_FTYPE',
    4 : 'WND4_WRITE_DELEG_NOT_SUPP_FTYPE',
    5 : 'WND4_NOT_SUPP_UPGRADE',
    6 : 'WND4_NOT_SUPP_DOWNGRADE',
    7 : 'WND4_CANCELED',
    8 : 'WND4_IS_DIR',
}
OPEN4_RESULT_CONFIRM = 0x00000002L
OPEN4_RESULT_LOCKTYPE_POSIX = 0x00000004L
OPEN4_RESULT_PRESERVE_UNLINKED = 0x00000008L
OPEN4_RESULT_MAY_NOTIFY_LOCK = 0x00000020L
RPC_GSS_SVC_NONE = 1
RPC_GSS_SVC_INTEGRITY = 2
RPC_GSS_SVC_PRIVACY = 3
rpc_gss_svc_t = {
    1 : 'RPC_GSS_SVC_NONE',
    2 : 'RPC_GSS_SVC_INTEGRITY',
    3 : 'RPC_GSS_SVC_PRIVACY',
}
UNSTABLE4 = 0
DATA_SYNC4 = 1
FILE_SYNC4 = 2
stable_how4 = {
    0 : 'UNSTABLE4',
    1 : 'DATA_SYNC4',
    2 : 'FILE_SYNC4',
}
CDFC4_FORE = 0x1
CDFC4_BACK = 0x2
CDFC4_FORE_OR_BOTH = 0x3
CDFC4_BACK_OR_BOTH = 0x7
channel_dir_from_client4 = {
    0x1 : 'CDFC4_FORE',
    0x2 : 'CDFC4_BACK',
    0x3 : 'CDFC4_FORE_OR_BOTH',
    0x7 : 'CDFC4_BACK_OR_BOTH',
}
CDFS4_FORE = 0x1
CDFS4_BACK = 0x2
CDFS4_BOTH = 0x3
channel_dir_from_server4 = {
    0x1 : 'CDFS4_FORE',
    0x2 : 'CDFS4_BACK',
    0x3 : 'CDFS4_BOTH',
}
EXCHGID4_FLAG_SUPP_MOVED_REFER = 0x00000001L
EXCHGID4_FLAG_SUPP_MOVED_MIGR = 0x00000002L
EXCHGID4_FLAG_BIND_PRINC_STATEID = 0x00000100L
EXCHGID4_FLAG_USE_NON_PNFS = 0x00010000L
EXCHGID4_FLAG_USE_PNFS_MDS = 0x00020000L
EXCHGID4_FLAG_USE_PNFS_DS = 0x00040000L
EXCHGID4_FLAG_MASK_PNFS = 0x00070000L
EXCHGID4_FLAG_UPD_CONFIRMED_REC_A = 0x40000000L
EXCHGID4_FLAG_CONFIRMED_R = 0x80000000L
SP4_NONE = 0
SP4_MACH_CRED = 1
SP4_SSV = 2
state_protect_how4 = {
    0 : 'SP4_NONE',
    1 : 'SP4_MACH_CRED',
    2 : 'SP4_SSV',
}
CREATE_SESSION4_FLAG_PERSIST = 0x00000001L
CREATE_SESSION4_FLAG_CONN_BACK_CHAN = 0x00000002L
CREATE_SESSION4_FLAG_CONN_RDMA = 0x00000004L
GDD4_OK = 0
GDD4_UNAVAIL = 1
gddrnf4_status = {
    0 : 'GDD4_OK',
    1 : 'GDD4_UNAVAIL',
}
SECINFO_STYLE4_CURRENT_FH = 0
SECINFO_STYLE4_PARENT = 1
secinfo_style4 = {
    0 : 'SECINFO_STYLE4_CURRENT_FH',
    1 : 'SECINFO_STYLE4_PARENT',
}
SEQ4_STATUS_CB_PATH_DOWN = 0x00000001L
SEQ4_STATUS_CB_GSS_CONTEXTS_EXPIRING = 0x00000002L
SEQ4_STATUS_CB_GSS_CONTEXTS_EXPIRED = 0x00000004L
SEQ4_STATUS_EXPIRED_ALL_STATE_REVOKED = 0x00000008L
SEQ4_STATUS_EXPIRED_SOME_STATE_REVOKED = 0x00000010L
SEQ4_STATUS_ADMIN_STATE_REVOKED = 0x00000020L
SEQ4_STATUS_RECALLABLE_STATE_REVOKED = 0x00000040L
SEQ4_STATUS_LEASE_MOVED = 0x00000080L
SEQ4_STATUS_RESTART_RECLAIM_NEEDED = 0x00000100L
SEQ4_STATUS_CB_PATH_DOWN_SESSION = 0x00000200L
SEQ4_STATUS_BACKCHANNEL_FAULT = 0x00000400L
SEQ4_STATUS_DEVID_CHANGED = 0x00000800L
SEQ4_STATUS_DEVID_DELETED = 0x00001000L
OP_ACCESS = 3
OP_CLOSE = 4
OP_COMMIT = 5
OP_CREATE = 6
OP_DELEGPURGE = 7
OP_DELEGRETURN = 8
OP_GETATTR = 9
OP_GETFH = 10
OP_LINK = 11
OP_LOCK = 12
OP_LOCKT = 13
OP_LOCKU = 14
OP_LOOKUP = 15
OP_LOOKUPP = 16
OP_NVERIFY = 17
OP_OPEN = 18
OP_OPENATTR = 19
OP_OPEN_CONFIRM = 20
OP_OPEN_DOWNGRADE = 21
OP_PUTFH = 22
OP_PUTPUBFH = 23
OP_PUTROOTFH = 24
OP_READ = 25
OP_READDIR = 26
OP_READLINK = 27
OP_REMOVE = 28
OP_RENAME = 29
OP_RENEW = 30
OP_RESTOREFH = 31
OP_SAVEFH = 32
OP_SECINFO = 33
OP_SETATTR = 34
OP_SETCLIENTID = 35
OP_SETCLIENTID_CONFIRM = 36
OP_VERIFY = 37
OP_WRITE = 38
OP_RELEASE_LOCKOWNER = 39
OP_BACKCHANNEL_CTL = 40
OP_BIND_CONN_TO_SESSION = 41
OP_EXCHANGE_ID = 42
OP_CREATE_SESSION = 43
OP_DESTROY_SESSION = 44
OP_FREE_STATEID = 45
OP_GET_DIR_DELEGATION = 46
OP_GETDEVICEINFO = 47
OP_GETDEVICELIST = 48
OP_LAYOUTCOMMIT = 49
OP_LAYOUTGET = 50
OP_LAYOUTRETURN = 51
OP_SECINFO_NO_NAME = 52
OP_SEQUENCE = 53
OP_SET_SSV = 54
OP_TEST_STATEID = 55
OP_WANT_DELEGATION = 56
OP_DESTROY_CLIENTID = 57
OP_RECLAIM_COMPLETE = 58
OP_ILLEGAL = 10044
nfs_opnum4 = {
    3 : 'OP_ACCESS',
    4 : 'OP_CLOSE',
    5 : 'OP_COMMIT',
    6 : 'OP_CREATE',
    7 : 'OP_DELEGPURGE',
    8 : 'OP_DELEGRETURN',
    9 : 'OP_GETATTR',
    10 : 'OP_GETFH',
    11 : 'OP_LINK',
    12 : 'OP_LOCK',
    13 : 'OP_LOCKT',
    14 : 'OP_LOCKU',
    15 : 'OP_LOOKUP',
    16 : 'OP_LOOKUPP',
    17 : 'OP_NVERIFY',
    18 : 'OP_OPEN',
    19 : 'OP_OPENATTR',
    20 : 'OP_OPEN_CONFIRM',
    21 : 'OP_OPEN_DOWNGRADE',
    22 : 'OP_PUTFH',
    23 : 'OP_PUTPUBFH',
    24 : 'OP_PUTROOTFH',
    25 : 'OP_READ',
    26 : 'OP_READDIR',
    27 : 'OP_READLINK',
    28 : 'OP_REMOVE',
    29 : 'OP_RENAME',
    30 : 'OP_RENEW',
    31 : 'OP_RESTOREFH',
    32 : 'OP_SAVEFH',
    33 : 'OP_SECINFO',
    34 : 'OP_SETATTR',
    35 : 'OP_SETCLIENTID',
    36 : 'OP_SETCLIENTID_CONFIRM',
    37 : 'OP_VERIFY',
    38 : 'OP_WRITE',
    39 : 'OP_RELEASE_LOCKOWNER',
    40 : 'OP_BACKCHANNEL_CTL',
    41 : 'OP_BIND_CONN_TO_SESSION',
    42 : 'OP_EXCHANGE_ID',
    43 : 'OP_CREATE_SESSION',
    44 : 'OP_DESTROY_SESSION',
    45 : 'OP_FREE_STATEID',
    46 : 'OP_GET_DIR_DELEGATION',
    47 : 'OP_GETDEVICEINFO',
    48 : 'OP_GETDEVICELIST',
    49 : 'OP_LAYOUTCOMMIT',
    50 : 'OP_LAYOUTGET',
    51 : 'OP_LAYOUTRETURN',
    52 : 'OP_SECINFO_NO_NAME',
    53 : 'OP_SEQUENCE',
    54 : 'OP_SET_SSV',
    55 : 'OP_TEST_STATEID',
    56 : 'OP_WANT_DELEGATION',
    57 : 'OP_DESTROY_CLIENTID',
    58 : 'OP_RECLAIM_COMPLETE',
    10044 : 'OP_ILLEGAL',
}
NFS4_PROGRAM = 100003
NFS_V4 = 4
LAYOUTRECALL4_FILE = LAYOUT4_RET_REC_FILE
LAYOUTRECALL4_FSID = LAYOUT4_RET_REC_FSID
LAYOUTRECALL4_ALL = LAYOUT4_RET_REC_ALL
layoutrecall_type4 = {
    LAYOUT4_RET_REC_FILE : 'LAYOUTRECALL4_FILE',
    LAYOUT4_RET_REC_FSID : 'LAYOUTRECALL4_FSID',
    LAYOUT4_RET_REC_ALL : 'LAYOUTRECALL4_ALL',
}
NOTIFY4_CHANGE_CHILD_ATTRS = 0
NOTIFY4_CHANGE_DIR_ATTRS = 1
NOTIFY4_REMOVE_ENTRY = 2
NOTIFY4_ADD_ENTRY = 3
NOTIFY4_RENAME_ENTRY = 4
NOTIFY4_CHANGE_COOKIE_VERIFIER = 5
notify_type4 = {
    0 : 'NOTIFY4_CHANGE_CHILD_ATTRS',
    1 : 'NOTIFY4_CHANGE_DIR_ATTRS',
    2 : 'NOTIFY4_REMOVE_ENTRY',
    3 : 'NOTIFY4_ADD_ENTRY',
    4 : 'NOTIFY4_RENAME_ENTRY',
    5 : 'NOTIFY4_CHANGE_COOKIE_VERIFIER',
}
RCA4_TYPE_MASK_RDATA_DLG = 0
RCA4_TYPE_MASK_WDATA_DLG = 1
RCA4_TYPE_MASK_DIR_DLG = 2
RCA4_TYPE_MASK_FILE_LAYOUT = 3
RCA4_TYPE_MASK_BLK_LAYOUT_MIN = 4
RCA4_TYPE_MASK_BLK_LAYOUT_MAX = 7
RCA4_TYPE_MASK_OBJ_LAYOUT_MIN = 8
RCA4_TYPE_MASK_OBJ_LAYOUT_MAX = 11
RCA4_TYPE_MASK_OTHER_LAYOUT_MIN = 12
RCA4_TYPE_MASK_OTHER_LAYOUT_MAX = 15
NOTIFY_DEVICEID4_CHANGE = 1
NOTIFY_DEVICEID4_DELETE = 2
notify_deviceid_type4 = {
    1 : 'NOTIFY_DEVICEID4_CHANGE',
    2 : 'NOTIFY_DEVICEID4_DELETE',
}
OP_CB_GETATTR = 3
OP_CB_RECALL = 4
OP_CB_LAYOUTRECALL = 5
OP_CB_NOTIFY = 6
OP_CB_PUSH_DELEG = 7
OP_CB_RECALL_ANY = 8
OP_CB_RECALLABLE_OBJ_AVAIL = 9
OP_CB_RECALL_SLOT = 10
OP_CB_SEQUENCE = 11
OP_CB_WANTS_CANCELLED = 12
OP_CB_NOTIFY_LOCK = 13
OP_CB_NOTIFY_DEVICEID = 14
OP_CB_ILLEGAL = 10044
nfs_cb_opnum4 = {
    3 : 'OP_CB_GETATTR',
    4 : 'OP_CB_RECALL',
    5 : 'OP_CB_LAYOUTRECALL',
    6 : 'OP_CB_NOTIFY',
    7 : 'OP_CB_PUSH_DELEG',
    8 : 'OP_CB_RECALL_ANY',
    9 : 'OP_CB_RECALLABLE_OBJ_AVAIL',
    10 : 'OP_CB_RECALL_SLOT',
    11 : 'OP_CB_SEQUENCE',
    12 : 'OP_CB_WANTS_CANCELLED',
    13 : 'OP_CB_NOTIFY_LOCK',
    14 : 'OP_CB_NOTIFY_DEVICEID',
    10044 : 'OP_CB_ILLEGAL',
}
NFS4_CALLBACK = 0x40000000L
NFS_CB = 1



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Unpack module

Provides the object for managing and unpacking raw data from a working buffer.
"""
import struct
import nfstest_config as c

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"

class Unpack(object):
    """Unpack object

       Usage:
           from packet.unpack import Unpack

           x = Unpack(buffer)

           # Get the 32 bytes from the working buffer
           data = x.rawdata(32)

           # Unpack an 'unsigned short' (2 bytes)
           short_int = x.unpack(2, 'H')[0]
    """
    def __init__(self, data):
        """Constructor

           Initialize object's private data.

           data:
               Raw packet data
        """
        self.data = data

    def rawdata(self, size):
        """Get the number of bytes given from the working buffer."""
        buf = self.data[0:size]
        self.data = self.data[size:]
        return buf

    def unpack(self, size, fmt):
        """Get the number of bytes given from the working buffer and process
           it according to the given format.
           Return a tuple of unpack items, see struct.unpack.
        """
        return struct.unpack('!'+fmt, self.rawdata(size))




#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
TCP module

Decode TCP layer.
"""
import nfstest_config as c
from baseobj import BaseObj
from packet.application.rpc import RPC
from packet.unpack import Unpack

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"

_TCP_map = {
    0x01:'FIN',
    0x02:'SYN',
    0x04:'RST',
    0x08:'PSH',
    0x10:'ACK',
    0x20:'URG',
    0x40:'ECE',
    0x80:'CWR',
}

class Flags(BaseObj): pass

class TCP(BaseObj, Unpack):
    """TCP object

       Usage:
           from packet.transport.tcp import TCP

           x = TCP(pktt, buffer)

       Object definition:

       TCP(
           src_port    = int,
           dst_port    = int,
           seq_number  = int,
           seq         = int, # relative sequence number
           ack_number  = int,
           hl          = int,
           header_size = int,
           window_size = int,
           checksum    = int,
           urgent_ptr  = int,
           flags_raw   = int, # raw flags
           flags = Flags(
               FIN = int,
               SYN = int,
               RST = int,
               PSH = int,
               ACK = int,
               URG = int,
               ECE = int,
               CWR = int,
           ),
           options = string, # raw data of TCP options if available
           data = string,    # raw data of payload if unable to decode
       )
    """
    def __init__(self, pktt, data):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
           data:
               Raw packet data for this layer.
        """
        self.data = data
        # Decode the TCP layer header
        ulist = self.unpack(20, 'HHIIBBHHH')
        temp = ulist[4] >> 4
        count = 4*temp
        self.src_port    = ulist[0]
        self.dst_port    = ulist[1]
        self.seq_number  = ulist[2]
        self.ack_number  = ulist[3]
        self.hl          = temp
        self.header_size = count
        self.window_size = ulist[6]
        self.checksum    = ulist[7]
        self.urgent_ptr  = ulist[8]
        self.flags_raw   = (ulist[5] & 0xFF)
        self.flags = Flags(
            FIN = (ulist[5] & 0x01),
            SYN = ((ulist[5] >> 1) & 0x01),
            RST = ((ulist[5] >> 2) & 0x01),
            PSH = ((ulist[5] >> 3) & 0x01),
            ACK = ((ulist[5] >> 4) & 0x01),
            URG = ((ulist[5] >> 5) & 0x01),
            ECE = ((ulist[5] >> 6) & 0x01),
            CWR = ((ulist[5] >> 7) & 0x01),
        )
        pktt.pkt.tcp = self

        # Stream identifier
        streamid = self._streamid(pktt.pkt)

        if not getattr(pktt, '_tcp_stream_map', None):
            # TCP stream map: to keep track of the different TCP streams
            # within the trace file -- used to deal with RPC packets spanning
            # multiple TCP packets or to handle a TCP packet having multiple
            # RPC packets
            pktt._tcp_stream_map = {}

        if streamid not in pktt._tcp_stream_map:
            # msfrag: Keep track of RPC packets spanning multiple TCP packets
            # frag_off: Keep track of multiple RPC packets within
            #           a single TCP packet
            pktt._tcp_stream_map[streamid] = {
                'seq_base': self.seq_number,
                'smap':     {},
                'pindex':   pktt.index,
                'msfrag':   '',
                'frag_off': 0,
                'last_seq': 0,
            }

        # De-reference stream map
        stream = pktt._tcp_stream_map[streamid]

        # Convert sequence numbers to relative numbers
        seq = self.seq_number - stream['seq_base']
        self.seq = seq

        if count > 20:
            osize = count - 20
            self.options = self.rawdata(osize)

        # Save length of TCP segment
        self.length = len(self.data)

        if seq < stream['last_seq']:
            # This is a re-transmission, do not process
            return

        # Save data
        save_data = self.data

        # Make sure this segment has valid data
        if len(self.data) <= 20 and self.data == '\x00' * len(self.data):
            save_data = ""

        # Append segment to the stream map
        if pktt.index == pktt.mindex:
            smap = stream['smap']
            if len(stream['msfrag']) == 0 and stream['frag_off'] == 0:
                stream['pindex'] = pktt.index
            else:
                smap_item = [stream['pindex'], stream['frag_off']]
                smap[pktt.index] = smap_item

        self._decode_payload(pktt, stream)

        if getattr(pktt.pkt, 'rpc', None) or len(save_data) == 0:
            stream['pindex'] = pktt.index

        if self.length > 0:
            stream['last_seq'] = seq
        return

    def __str__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned.
           If set to 1 the representation of the object is condensed:
               'TCP 708 -> 2049, seq: 3294175829, ack: 3395739041, ACK,FIN'

           If set to 2 the representation of the object also includes the
           length of payload and a little bit more verbose:
               'src port 708 -> dst port 2049, seq: 3294175829, ack: 3395739041, len: 0, flags: ACK,FIN'
        """
        rdebug = self.debug_repr()
        if rdebug > 0:
            flags = []
            for flag in _TCP_map:
                if self.flags_raw & flag != 0:
                    flags.append(_TCP_map[flag])
        if rdebug == 1:
            out = "TCP %d -> %d, seq: %d, ack: %d, %s" % \
                  (self.src_port, self.dst_port, self.seq_number, self.ack_number, ','.join(flags))
        elif rdebug == 2:
            out = "src port %d -> dst port %d, seq: %d, ack: %d, len: %d, flags: %s" % \
                  (self.src_port, self.dst_port, self.seq_number, self.ack_number, self.length, ','.join(flags))
        else:
            out = BaseObj.__str__(self)
        return out

    def _decode_payload(self, pktt, stream):
        """Decode TCP payload."""
        rpc = None
        if stream['frag_off'] > 0 and len(stream['msfrag']) == 0:
            # This RPC packet lies within previous TCP packet,
            # Re-position the offset of the data
            self.data = self.data[stream['frag_off']:]

        # Get the total size
        save_data = self.data
        size = len(self.data)

        # Try decoding the RPC header before using the msfrag data
        # to re-sync the stream
        if len(stream['msfrag']) > 0:
            rpc = RPC(pktt, self.data, proto=6)
            if not rpc:
                self.data = save_data

        if rpc or (size == 0 and len(stream['msfrag']) > 0 and self.flags_raw != 0x10):
            # There has been some data lost in the capture,
            # to continue decoding next packets, reset stream
            # except if this packet is just a TCP ACK (flags = 0x10)
            stream['msfrag'] = ''
            stream['frag_off'] = 0

        # Make sure this segment has valid data
        if size <= 20 and save_data == '\x00' * size:
            return

        if not rpc:
            # Concatenate previous fragment
            self.data = stream['msfrag'] + self.data
            ldata = len(self.data) - 4

            # Get RPC header
            rpc = RPC(pktt, self.data, proto=6)
        else:
            ldata = size - 4

        if not rpc:
            return

        rpcsize = rpc.fragment_hdr.size

        if ldata < rpcsize:
            # An RPC fragment is missing to decode RPC payload
            stream['msfrag'] += save_data
        else:
            if len(stream['msfrag']) > 0 or ldata == rpcsize:
                stream['frag_off'] = 0
            stream['msfrag'] = ''
            # Save RPC layer on packet object
            pktt.pkt.rpc = rpc
            del self.data

            # Decode NFS layer
            nfs = rpc.decode_nfs()
            if nfs:
                pktt.pkt.nfs = nfs
            rpcbytes = ldata - len(rpc.data)
            if not nfs and rpcbytes != rpcsize:
                pass
            elif rpc.data:
                # Save the offset of next RPC packet within this TCP packet
                # Data offset is cumulative
                stream['frag_off'] += size - len(rpc.data)
                save_data = rpc.data
                ldata = len(rpc.data) - 4
                try:
                    rpc_header = RPC(pktt, rpc.data, proto=6)
                except Exception:
                    rpc_header = None
                if not rpc_header or ldata < rpc_header.fragment_hdr.size:
                    # Part of next RPC packet is within this TCP packet
                    # Save the multi-span fragment data
                    stream['msfrag'] += save_data
                else:
                    # Next RPC packet is entirely within this TCP packet
                    # Re-position the file pointer to the current offset
                    pktt.offset = pktt.b_offset
                    pktt._getfh().seek(pktt.offset)
            else:
                stream['frag_off'] = 0

    def _streamid(self, pkt):
        """Get TCP streamid."""
        streamid = "%s:%d-%s:%d" % (pkt.ip.src,
                                    pkt.tcp.src_port,
                                    pkt.ip.dst,
                                    pkt.tcp.dst_port)
        return streamid




#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Base object

Base class so objects will inherit the methods which provide the string
representation of the object and methods to change the verbosity of such
string representation. It also includes a simple debug printing and logging
mechanism including methods to change the debug verbosity level and methods
to add debug levels.
"""
import re
import nfstest_config as c
from pprint import pformat,pprint

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__version__   = '1.0.1'
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"

# Module variables
_dindent = 0
_dlevel = 0
_rlevel = 0
_logfh = None
# Simple verbose level names
_debug_map = {
    'none':  0,
    'info':  1,    # Display info only
    'debug': 0xFF, # Display info and all debug messages 0x02-0x80
    'all':   0xFFFFFFFF, # Display all messages
}
# Debug display prefixes
_debug_prefix = {
    0x001: 'INFO: ',
}

def _init_debug():
    """Define all debug flags"""
    for i in xrange(7):
        dbg = 'dbg%d' % (i+1)
        _debug_map[dbg] = (2 << i)
        _debug_prefix[(2 << i)] = dbg.upper() + ': '
_init_debug()

class BaseObj(object):
    """Base class so objects will inherit the methods which provide the string
       representation of the object and a simple debug printing and logging
       mechanism.
    """
    def __init__(self, *kwts, **kwds):
        """Constructor

           Initialize object's private data according to the arguments given

           Examples:
               # Named arguments
               x = BaseObj(a=1, b=2)

               # Dictionary argument
               x = BaseObj({'a':1, 'b':2})

               # Tuple arguments: first for keys and second for the values
               x = BaseObj(['a', 'b'], [1, 2])

               # All of the above will create an object having two attributes:
               x.a = 1 and x.b = 2
        """
        keys = None
        for item in kwts:
            if type(item) == dict:
                self.__dict__.update(item)
            elif type(item) == list or type(item) == tuple:
                if keys is None:
                    keys = item
                else:
                    self.__dict__.update(zip(keys,item))
                    keys = None
        # Process named arguments: x = BaseObj(a=1, b=2)
        self.__dict__.update(kwds)

    def __repr__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned, else
           the representation of the object includes all object attributes
           and their values with proper indentation.
        """
        global _rlevel
        if _rlevel == 0:
            # Return generic object representation
            return object.__repr__(self)

        # Representation of object with proper indentation
        indent = ' ' * 4
        out = self.__class__.__name__ + "(\n"
        itemlist = getattr(self, '_itemlist', None)
        if itemlist is None:
            itemlist = sorted(self.__dict__.iterkeys())
        for key in itemlist:
            if key[0] != '_':
                val = self.__dict__.get(key, None)
                if val != None:
                    value = pformat(val)
                    out += "    %s = %s,\n" % (key, value.replace("\n", "\n"+indent))
                else:
                    out += "    %s = None,\n" % key
        out += ")"
        return out
    __str__ = __repr__

    @staticmethod
    def debug_repr(level=None):
        """Return or set verbose level of object's string representation.
           When setting the verbose level, return the verbose level before
           setting it.

           level:
               Level of verbosity to set
        """
        global _rlevel
        ret = _rlevel
        if level is not None:
            _rlevel = level
        return ret

    def debug_level(self, level=0):
        """Set debug level mask.

           level:
               Level to set. This could be a number or a string expression
               of names defined by debug_map()

           Examples:
               # Set level
               x.debug_level(0xFF)

               # Set level using expression
               x.debug_level('all')
               x.debug_level('debug ^ 1')
        """
        global _dlevel
        if type(level) == str:
            # Convert named verbose levels to a number
            # -- Get a list of all named verbose levels
            for item in sorted(set(re.split('\W+', level))):
                if len(item) > 0:
                    if item in _debug_map:
                        # Replace all occurrences of named verbose level
                        # to its corresponding numeric value
                        level = re.sub(r'\b' + item + r'\b', hex(_debug_map[item]), level)
                    else:
                        try:
                            # Find out if verbose is a number
                            # (decimal, hex, octal, ...)
                            tmp = int(item, 0)
                        except:
                            raise Exception("Unknown debug level [%s]" % item)
            # Evaluate the whole expression
            _dlevel = eval(level)
        else:
            # Already a number
            _dlevel = level
        return _dlevel

    @staticmethod
    def debug_map(bitmap, name='', disp=''):
        """Add a debug mapping.

           Generic debug levels map
             <bitmap>  <name>  <disp prefix>
              0x000    'none'
              0x001    'info'  'INFO: ' # Display info messages only
              0x0FF    'debug' 'DBG:  ' # Display info and all debug messages (0x02-0x80)
             >0x100    user defined verbose levels
        """
        if name:
            _debug_map[name] = bitmap
        if disp:
            _debug_prefix[bitmap] = disp

    @staticmethod
    def dindent(indent):
        """Set global indentation."""
        global _dindent
        _dindent = indent

    def open_log(self, logfile):
        """Open log file."""
        global _logfh
        self.close_log()
        _logfh = open(logfile, "w")

    def close_log(self):
        """Close log file."""
        global _logfh
        if _logfh != None:
            _logfh.close()
            _logfh = None

    @staticmethod
    def write_log(data):
        """Write data to log file."""
        if _logfh != None:
            _logfh.write(data + "\n")

    def dprint(self, level, msg, indent=0):
        """Print debug message if level is allowed by the verbose level
           given in debug_level().
        """
        ret = ''
        if level is None:
            return
        if type(level) == str:
            level = _debug_map[level.lower()]
        if level & _dlevel:
            # Add display prefix only if msg is not an empty string
            if len(msg):
                # Find the right display prefix
                prefix = ' ' * _dindent
                for bitmap in sorted(_debug_prefix):
                    if level & bitmap:
                        prefix += _debug_prefix[bitmap]
                        break
                # Add display prefix to the message
                sp = ' ' * indent
                ret = prefix + sp + msg
                indent += len(prefix)
            if indent > 0:
                sp = ' ' * indent
                ret = ret.replace("\n", "\n"+sp)
            print ret
            self.write_log(ret)

